from typing import TYPE_CHECKING, Any

from sqlalchemy import Column, DateTime, ForeignKey, Index, String, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql import func

from fides.api.db.base_class import Base, FidesBase
from fides.api.db.util import EnumColumn
from fides.api.models.manual_tasks.manual_task_log import ManualTaskLog
from fides.api.schemas.manual_tasks.manual_task_schemas import (
    ManualTaskLogStatus,
    ManualTaskParentEntityType,
    ManualTaskReferenceType,
    ManualTaskType,
)

if TYPE_CHECKING:  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task_config import (  # pragma: no cover
        ManualTaskConfig,
    )
    from fides.api.models.manual_tasks.manual_task_instance import (
        ManualTaskInstance,
        ManualTaskSubmission,
    )


class ManualTask(Base):
    """Model for storing manual tasks.

    This model can be used for both privacy request tasks and general tasks.
    For privacy requests, it replaces the functionality of manual webhooks.
    For other use cases, it provides a flexible task management system.

    There can only be one ManualTask per parent entity.
    You can create multiple Configs for the same ManualTask.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_type = Column(
        EnumColumn(ManualTaskType),
        nullable=False,
        default=ManualTaskType.privacy_request,
    )
    parent_entity_id = Column(String, nullable=False)
    parent_entity_type = Column(
        EnumColumn(ManualTaskParentEntityType),
        nullable=False,
        default=ManualTaskParentEntityType.connection_config,
    )

    __table_args__ = (
        UniqueConstraint(
            "parent_entity_id",
            "parent_entity_type",
            name="uq_manual_task_parent_entity",
        ),
        Index("ix_manual_task_parent_entity", "parent_entity_type", "parent_entity_id"),
        Index("ix_manual_task_task_type", "task_type"),
    )

    # Relationships
    references = relationship(
        "ManualTaskReference",
        back_populates="task",
        uselist=True,
        cascade="all, delete-orphan",
    )
    logs = relationship(
        "ManualTaskLog",
        back_populates="task",
        primaryjoin="and_(ManualTask.id == ManualTaskLog.task_id)",
        viewonly=True,
        order_by="ManualTaskLog.created_at",
    )
    configs = relationship(
        "ManualTaskConfig",
        back_populates="task",
        cascade="all, delete-orphan",
        uselist=True,
    )
    instances = relationship(
        "ManualTaskInstance",
        back_populates="task",
        viewonly=True,
        uselist=True,
    )
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="task",
        uselist=True,
        viewonly=True,
    )

    # Properties
    @property
    def assigned_users(self) -> list[str]:
        """Get all users assigned to this task."""
        if not self.references:
            return []
        return [
            ref.reference_id
            for ref in self.references
            if ref.reference_type == ManualTaskReferenceType.assigned_user
        ]

    # CRUD Operations
    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = True
    ) -> "ManualTask":
        """Create a new manual task."""
        task = super().create(db=db, data=data, check_name=check_name)
        ManualTaskLog.create_log(
            db=db,
            task_id=task.id,
            status=ManualTaskLogStatus.created,
            message=f"Created manual task for {data['task_type']}",
        )
        return task


class ManualTaskReference(Base):
    """Join table to associate manual tasks with multiple references.

    A single task may have many references including privacy requests, configurations, and assigned users.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_reference"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_id = Column(
        String, ForeignKey("manual_task.id", ondelete="CASCADE"), nullable=False
    )
    reference_id = Column(String, nullable=False)
    reference_type = Column(EnumColumn(ManualTaskReferenceType), nullable=False)

    __table_args__ = (
        Index("ix_manual_task_reference_reference", "reference_id", "reference_type"),
        Index("ix_manual_task_reference_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="references", viewonly=True)



"""adding foreign key on fides key for taxonomy

Revision ID: 10c6b7709be3
Revises: b63ecb007556
Create Date: 2024-12-17 14:54:02.325442

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "10c6b7709be3"
down_revision = "b63ecb007556"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_foreign_key(
        "ctl_data_categories_parent_key_fkey",
        "ctl_data_categories",
        "ctl_data_categories",
        ["parent_key"],
        ["fides_key"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        "ctl_data_uses_parent_key_fkey",
        "ctl_data_uses",
        "ctl_data_uses",
        ["parent_key"],
        ["fides_key"],
        ondelete="RESTRICT",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "ctl_data_categories_parent_key_fkey", "ctl_data_categories", type_="foreignkey"
    )
    op.drop_constraint(
        "ctl_data_uses_parent_key_fkey", "ctl_data_uses", type_="foreignkey"
    )
    # ### end Alembic commands ###



"""Add finished audit log

Revision ID: bab75915670a
Revises: 3c5e1253465d
Create Date: 2022-08-04 20:18:29.339116

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bab75915670a"
down_revision = "3c5e1253465d"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type auditlogaction add value 'finished'")


def downgrade():
    op.execute("delete from auditlog where action in ('finished')")
    op.execute("alter type auditlogaction rename to auditlogaction_old")
    op.execute("create type auditlogaction as enum('approved', 'denied')")
    op.execute(
        (
            "alter table auditlog alter column action type auditlogaction using "
            "action::text::auditlogaction"
        )
    )
    op.execute("drop type auditlogaction_old")



import logging


class SQLAlchemyGeneratedFilter(logging.Filter):
    """
    Logging filter that removes SQLAlchemy engine logs containing timing information.

    This filter specifically targets logs containing error messages from sqlalchemy.engine.Engine
    which typically come from sqlalchemy.engine.Engine when executing cached queries.
    """

    def __init__(self) -> None:
        super().__init__()
        # List of substrings to filter out
        self.patterns = [
            "cached since",
            "caching disabled",
            "does not support caching",
        ]

    def filter(self, record: logging.LogRecord) -> bool:
        # Only filter records from sqlalchemy.engine.Engine
        if record.name == "sqlalchemy.engine.Engine":
            message = record.getMessage()
            # Check if any of the patterns exist in the message
            for pattern in self.patterns:
                if pattern in message:
                    return False  # Filter out the log entry

        return True  # Keep all other log entries



from typing import TYPE_CHECKING, Any, Optional

from sqlalchemy import Column, DateTime, ForeignKey, Index, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql import func

from fides.api.db.base_class import Base, FidesBase
from fides.api.schemas.manual_tasks.manual_task_schemas import ManualTaskLogStatus

if TYPE_CHECKING:  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task import ManualTask  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task_config import (
        ManualTaskConfig,  # pragma: no cover
    )
    from fides.api.models.manual_tasks.manual_task_instance import (  # pragma: no cover
        ManualTaskInstance,
    )


class ManualTaskLog(Base):
    """Model for storing manual task execution logs."""

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_log"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    task_id = Column(
        String, ForeignKey("manual_task.id", ondelete="CASCADE"), nullable=False
    )
    config_id = Column(
        String, ForeignKey("manual_task_config.id", ondelete="CASCADE"), nullable=True
    )
    instance_id = Column(
        String,
        ForeignKey("manual_task_instance.id", ondelete="CASCADE"),
        nullable=True,
    )
    status = Column(String, nullable=False)
    message = Column(String, nullable=True)
    details = Column(JSONB, nullable=True)

    __table_args__ = (
        Index("ix_manual_task_log_config_id", "config_id"),
        Index("ix_manual_task_log_created_at", "created_at"),
        Index("ix_manual_task_log_instance_id", "instance_id"),
        Index("ix_manual_task_log_status", "status"),
        Index("ix_manual_task_log_task_id", "task_id"),
    )

    # Relationships
    task = relationship(
        "ManualTask", back_populates="logs", foreign_keys=[task_id], viewonly=True
    )
    config = relationship(
        "ManualTaskConfig",
        back_populates="logs",
        foreign_keys=[config_id],
        viewonly=True,
    )
    instance = relationship("ManualTaskInstance", back_populates="logs", viewonly=True)

    @classmethod
    def create_log(
        cls,
        db: Session,
        task_id: str,
        status: ManualTaskLogStatus,
        message: str,
        config_id: Optional[str] = None,
        instance_id: Optional[str] = None,
        details: Optional[dict[str, Any]] = None,
    ) -> "ManualTaskLog":
        """Create a new task log entry.

        Args:
            db: Database session
            task_id: ID of the task
            status: Status of the log entry
            message: Optional message describing the event
            details: Optional additional details about the event
        """
        data = {
            "task_id": task_id,
            "config_id": config_id,
            "instance_id": instance_id,
            "status": status,
            "message": message,
            "details": details,
        }
        return cls.create(db=db, data=data)

    @classmethod
    def create_error_log(
        cls,
        db: Session,
        task_id: str,
        message: str,
        config_id: Optional[str] = None,
        instance_id: Optional[str] = None,
        details: Optional[dict[str, Any]] = None,
    ) -> "ManualTaskLog":
        """Create a new error log entry.

        Args:
            db: Database session
            task_id: ID of the task
            message: Error message describing what went wrong
            config_id: Optional ID of the configuration
            instance_id: Optional ID of the instance
            details: Optional additional details about the error

        Returns:
            The created error log entry
        """
        return cls.create_log(
            db=db,
            status=ManualTaskLogStatus.error,
            task_id=task_id,
            config_id=config_id,
            instance_id=instance_id,
            message=message,
            details=details,
        )



"""add application config table

Revision ID: 5d62bab40b71
Revises: 7fd4601ef88b
Create Date: 2023-01-27 22:56:33.367379

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "5d62bab40b71"
down_revision = "7fd4601ef88b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "applicationconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "api_set",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("single_row", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.CheckConstraint("single_row", name="single_row_check"),
    )
    op.create_index(
        op.f("ix_applicationconfig_id"), "applicationconfig", ["id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_applicationconfig_id"), table_name="applicationconfig")
    op.drop_table("applicationconfig")
    # ### end Alembic commands ###



"""add default policies

Revision ID: 55d61eb8ed12
Revises: b3b68c87c4a0
Create Date: 2022-06-13 19:26:24.197262

"""

import logging

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

revision = "55d61eb8ed12"
down_revision = "b3b68c87c4a0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Data migration only."""

    # These data migrations have been removed, as they are now handled by
    # the `seed.py` module.
    pass


def downgrade() -> None:
    """Data migration only."""

    pass



from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Annotated, Literal, Optional, Union

from pydantic import ConfigDict, Field, ValidationInfo, field_validator, model_validator

from fides.api.schemas.base_class import FidesSchema

if TYPE_CHECKING:  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task import ManualTask  # pragma: no cover


class ManualTaskConfigurationType(str, Enum):
    """Enum for manual task configuration types."""

    access_privacy_request = "access_privacy_request"
    erasure_privacy_request = "erasure_privacy_request"
    # Add more configuration types as needed


class ManualTaskFieldType(str, Enum):
    """Enum for manual task field types."""

    text = "text"  # Key-value pairs
    checkbox = "checkbox"  # Boolean value
    attachment = "attachment"  # File upload
    # Add more field types as needed


class ManualTaskFieldMetadata(FidesSchema):
    """Base schema for manual task field metadata."""

    model_config = ConfigDict(extra="allow")

    label: Annotated[str, Field(description="Display label for the field")]
    required: Annotated[
        bool, Field(default=False, description="Whether the field is required")
    ]
    help_text: Annotated[
        Optional[str],
        Field(default=None, description="Help text to display with the field"),
    ]
    data_uses: Annotated[
        Optional[list[str]],
        Field(
            default=None,
            description="List of data uses associated with this field",
        ),
    ]


class ManualTaskTextFieldMetadata(ManualTaskFieldMetadata):
    """Schema text field metadata."""

    placeholder: Annotated[
        Optional[str], Field(default=None, description="Placeholder text for the field")
    ]
    default_value: Annotated[
        Optional[str], Field(default=None, description="Default value for the field")
    ]
    max_length: Annotated[
        Optional[int], Field(default=None, description="Maximum length for text fields")
    ]
    min_length: Annotated[
        Optional[int], Field(default=None, description="Minimum length for text fields")
    ]
    pattern: Annotated[
        Optional[str],
        Field(
            default=None,
            description=(
                "Regex pattern for validation, for example if the field is an email address, "
                r"you could use: ^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
            ),
        ),
    ]

    @field_validator("min_length", "max_length")
    @classmethod
    def validate_lengths(cls, v: Optional[int], info: ValidationInfo) -> Optional[int]:
        """Validate that min_length is not greater than max_length."""
        if v is None:
            return v

        if info.field_name == "min_length":
            max_length = info.data.get("max_length")
            if max_length is not None and v > max_length:
                raise ValueError("min_length cannot be greater than max_length")
        elif info.field_name == "max_length":
            min_length = info.data.get("min_length")
            if min_length is not None and v < min_length:
                raise ValueError("min_length cannot be greater than max_length")

        return v


class ManualTaskCheckboxFieldMetadata(ManualTaskFieldMetadata):
    """Schema for checkbox field metadata."""

    default_value: Annotated[
        Optional[bool],
        Field(default=False, description="Default value for the checkbox"),
    ]
    label_true: Annotated[
        Optional[str], Field(default=None, description="Label for checked state")
    ]
    label_false: Annotated[
        Optional[str], Field(default=None, description="Label for unchecked state")
    ]


class ManualTaskAttachmentFieldMetadata(ManualTaskFieldMetadata):
    """Schema for attachment field metadata."""

    file_types: Annotated[
        list[str], Field(description="Allowed file types for upload fields")
    ]
    max_file_size: Annotated[
        int, Field(description="Maximum file size in bytes for upload fields")
    ]
    multiple: Annotated[
        bool, Field(default=False, description="Whether multiple files can be uploaded")
    ]
    max_files: Annotated[
        Optional[int],
        Field(
            default=None,
            description="Maximum number of files allowed when multiple is true",
        ),
    ]
    require_attachment_id: Annotated[
        bool,
        Field(
            default=True,
            description="Whether an attachment ID is required for this field",
        ),
    ]
    attachment_ids: Annotated[
        Optional[list[str]],
        Field(
            default=None,
            description="List of attachment IDs associated with this field",
        ),
    ]

    @model_validator(mode="after")
    def validate_multiple_max_files(self) -> "ManualTaskAttachmentFieldMetadata":
        """Validate that max_files is set when multiple is True."""
        if self.multiple and self.max_files is None:
            raise ValueError("max_files must be set when multiple is True")
        return self


class ManualTaskFieldBase(FidesSchema):
    """Base schema for all field definitions."""

    model_config = ConfigDict(extra="allow")

    field_key: Annotated[
        str, Field(description="Unique key for the field", min_length=1)
    ]
    field_type: Annotated[ManualTaskFieldType, Field(description="Type of the field")]
    field_metadata: Annotated[
        ManualTaskFieldMetadata, Field(description="Field metadata and configuration")
    ]

    @classmethod
    def get_field_model_for_type(
        cls, field_type: Union[str, ManualTaskFieldType]
    ) -> type["ManualTaskFieldBase"]:
        """Get the appropriate field model class for a given field type.

        Args:
            field_type: The field type as a string or ManualTaskFieldType enum value

        Returns:
            The appropriate field model class

        Raises:
            ValueError: If the field type is not recognized
        """
        # Handle non-string/non-enum types
        if not isinstance(field_type, (str, ManualTaskFieldType)):
            raise ValueError(
                f"Invalid field type: expected string or ManualTaskFieldType, got {type(field_type).__name__}"
            )

        # Convert string to enum if needed
        if isinstance(field_type, str):
            try:
                field_type = ManualTaskFieldType(field_type)
            except ValueError:
                raise ValueError(
                    f"Invalid field type: '{field_type}' is not a valid ManualTaskFieldType"
                )

        # Map field types to their model classes
        field_type_to_model = {
            ManualTaskFieldType.text: ManualTaskTextField,
            ManualTaskFieldType.checkbox: ManualTaskCheckboxField,
            ManualTaskFieldType.attachment: ManualTaskAttachmentField,
        }

        model_class = field_type_to_model.get(field_type)
        if not model_class:
            raise ValueError(f"No model class found for field type: {field_type}")

        return model_class


class ManualTaskTextField(ManualTaskFieldBase):
    """Schema for text field definition."""

    field_type: Annotated[
        Literal[ManualTaskFieldType.text], Field(description="Type of the field")
    ]
    field_metadata: Annotated[
        ManualTaskTextFieldMetadata,
        Field(description="Field metadata and configuration"),
    ]


class ManualTaskCheckboxField(ManualTaskFieldBase):
    """Schema for checkbox field definition."""

    field_type: Annotated[
        Literal[ManualTaskFieldType.checkbox], Field(description="Type of the field")
    ]
    field_metadata: Annotated[
        ManualTaskCheckboxFieldMetadata,
        Field(description="Field metadata and configuration"),
    ]


class ManualTaskAttachmentField(ManualTaskFieldBase):
    """Schema for attachment field definition."""

    field_type: Annotated[
        Literal[ManualTaskFieldType.attachment], Field(description="Type of the field")
    ]
    field_metadata: Annotated[
        ManualTaskAttachmentFieldMetadata,
        Field(description="Field metadata and configuration"),
    ]


# Type alias for all field types
ManualTaskField = ManualTaskFieldBase


class ManualTaskConfigCreate(FidesSchema):
    """Schema for creating a manual task configuration."""

    model_config = ConfigDict(extra="forbid")

    task_id: Annotated[
        str, Field(description="ID of the task this configuration belongs to")
    ]
    config_type: Annotated[
        ManualTaskConfigurationType, Field(description="Type of configuration")
    ]
    fields: Annotated[
        list[ManualTaskFieldBase], Field(description="List of field definitions")
    ]


class ManualTaskConfigResponse(FidesSchema):
    """Schema for manual task configuration response."""

    model_config = ConfigDict(extra="forbid")

    id: Annotated[str, Field(description="Configuration ID")]
    task_id: Annotated[str, Field(description="Task ID")]
    config_type: Annotated[
        ManualTaskConfigurationType, Field(description="Type of configuration")
    ]
    version: Annotated[int, Field(description="Version of the configuration")]
    is_current: Annotated[
        bool, Field(description="Whether this is the current version")
    ]
    fields: Annotated[
        list[ManualTaskFieldBase], Field(description="List of field definitions")
    ]
    created_at: Annotated[datetime, Field(description="Creation timestamp")]
    updated_at: Annotated[datetime, Field(description="Last update timestamp")]


class ManualTaskConfigUpdate(FidesSchema):
    """Schema for updating a manual task configuration."""

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    task_id: Annotated[str, Field(description="ID of the task")]
    config_type: Annotated[
        ManualTaskConfigurationType, Field(description="Type of configuration")
    ]
    fields: Annotated[
        list[ManualTaskFieldBase], Field(description="List of field definitions")
    ]


class ManualTaskConfigFieldResponse(FidesSchema):
    """Schema for manual task configuration field response."""

    model_config = ConfigDict(extra="forbid")

    field_key: Annotated[str, Field(description="Unique key for the field")]
    field_type: Annotated[ManualTaskFieldType, Field(description="Type of the field")]
    field_metadata: Annotated[
        ManualTaskFieldMetadata, Field(description="Field metadata and configuration")
    ]



"""This module is responsible for parsing and verifying file, either with or without a server being available."""

from fideslang.manifests import ingest_manifests
from fideslang.models import Taxonomy
from fideslang.parse import load_manifests_into_taxonomy

from fides.common.utils import echo_green
from fides.core.utils import get_manifest_list


def parse(manifests_dir: str) -> Taxonomy:
    """
    Parse local manifest file(s) into a Taxonomy.
    """

    # Check if any manifests exist before trying to parse them
    print(f"Loading resource manifests from: {manifests_dir}")
    if not get_manifest_list(manifests_dir):
        print("No manifests found to parse, skipping...")
        return Taxonomy()
    ingested_manifests = ingest_manifests(manifests_dir)
    taxonomy = load_manifests_into_taxonomy(ingested_manifests)
    echo_green("Taxonomy successfully created.")
    return taxonomy



"""add shared monitor config

Revision ID: 2263583b0e44
Revises: d0cbfec0b2dd
Create Date: 2025-05-12 23:44:11.809581

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "2263583b0e44"
down_revision = "d0cbfec0b2dd"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "shared_monitor_config",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column(
            "classify_params",
            postgresql.JSONB(astext_type=sa.Text()),
            server_default="{}",
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("key"),
    )
    op.create_index(
        op.f("ix_shared_monitor_config_id"),
        "shared_monitor_config",
        ["id"],
        unique=False,
    )
    op.add_column(
        "monitorconfig", sa.Column("shared_config_id", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_monitorconfig_shared_config_id"),
        "monitorconfig",
        ["shared_config_id"],
        unique=False,
    )
    op.create_foreign_key(
        "fk_monitorconfig_shared_config_id",
        "monitorconfig",
        "shared_monitor_config",
        ["shared_config_id"],
        ["id"],
        ondelete="RESTRICT",
    )


def downgrade():
    op.drop_constraint(
        "fk_monitorconfig_shared_config_id", "monitorconfig", type_="foreignkey"
    )
    op.drop_index(op.f("ix_monitorconfig_shared_config_id"), table_name="monitorconfig")
    op.drop_column("monitorconfig", "shared_config_id")
    op.drop_index(
        op.f("ix_shared_monitor_config_id"), table_name="shared_monitor_config"
    )
    op.drop_table("shared_monitor_config")



from typing import Any, Dict, List, Optional

from loguru import logger
from sqlalchemy.orm import Query, Session

from fides.api.common_exceptions import MessageDispatchException
from fides.api.models.connectionconfig import (
    ConnectionConfig,
    ConnectionTestStatus,
    ConnectionType,
)
from fides.api.models.privacy_request import ExecutionLog
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.connection_configuration import EmailSchema
from fides.api.schemas.policy import ActionType
from fides.api.service.connectors.base_erasure_email_connector import (
    BaseErasureEmailConnector,
    filter_user_identities_for_connector,
    send_single_erasure_email,
)
from fides.config import get_config

CONFIG = get_config()

ERASURE_EMAIL_CONNECTOR_TYPES = [
    ConnectionType.generic_erasure_email,
    ConnectionType.attentive_email,
    ConnectionType.dynamic_erasure_email,
]


class GenericErasureEmailConnector(BaseErasureEmailConnector):
    """Generic Email Erasure Connector that can be overridden for specific vendors"""

    config: EmailSchema

    def get_config(self, configuration: ConnectionConfig) -> EmailSchema:
        return EmailSchema(**configuration.secrets or {})

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Sends an email to the "test_email" configured, just to establish that the email workflow is working.
        """
        logger.info("Starting test connection to {}", self.configuration.key)

        db = Session.object_session(self.configuration)

        try:
            if not self.config.test_email_address:
                raise MessageDispatchException(
                    f"Cannot test connection. No test email defined for {self.configuration.key}"
                )
            # synchronous for now since failure to send is considered a connection test failure
            send_single_erasure_email(
                db=db,
                subject_email=self.config.test_email_address,
                subject_name=self.config.third_party_vendor_name,
                batch_identities=list(self.identities_for_test_email.values()),
                test_mode=True,
            )
        except MessageDispatchException as exc:
            logger.info(
                "Email connector test for {} failed with exception {}",
                self.configuration.key,
                exc,
            )
            return ConnectionTestStatus.failed
        return ConnectionTestStatus.succeeded

    def batch_email_send(self, privacy_requests: Query) -> None:
        skipped_privacy_requests: List[str] = []
        batched_identities: List[str] = []
        db = Session.object_session(self.configuration)

        for privacy_request in privacy_requests:
            user_identities: Dict[str, Any] = privacy_request.get_cached_identity_data()
            filtered_user_identities: Dict[str, Any] = (
                filter_user_identities_for_connector(self.config, user_identities)
            )
            if filtered_user_identities:
                batched_identities.extend(filtered_user_identities.values())
            else:
                skipped_privacy_requests.append(privacy_request.id)
                self.add_skipped_log(db, privacy_request)

        if not batched_identities:
            logger.info(
                "Skipping erasure email send for connector: '{}'. "
                "No corresponding user identities found for pending privacy requests.",
                self.configuration.key,
            )
            return

        logger.info(
            "Sending batched erasure email for connector {}...",
            self.configuration.key,
        )

        try:
            send_single_erasure_email(
                db=db,
                subject_email=self.config.recipient_email_address,
                subject_name=self.config.third_party_vendor_name,
                batch_identities=batched_identities,
                test_mode=False,
            )
        except MessageDispatchException as exc:
            logger.info(
                "Erasure email for connector {} failed with exception {}",
                self.configuration.key,
                exc,
            )
            raise

        # create an audit event for each privacy request ID
        for privacy_request in privacy_requests:
            if privacy_request.id not in skipped_privacy_requests:
                ExecutionLog.create(
                    db=db,
                    data={
                        "connection_key": self.configuration.key,
                        "dataset_name": self.configuration.name_or_key,
                        "collection_name": self.configuration.name_or_key,
                        "privacy_request_id": privacy_request.id,
                        "action_type": ActionType.erasure,
                        "status": ExecutionLogStatus.complete,
                        "message": f"Erasure email instructions dispatched for '{self.configuration.name_or_key}'",
                    },
                )



from typing import Any, Dict, List, Optional

from loguru import logger

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.connection_configuration.connection_secrets_s3 import S3Schema
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.aws_util import get_aws_session
from fides.api.util.collection_util import Row


class S3Connector(BaseConnector):
    """
    AWS S3 Connector - this is currently used just to test connections to S3.

    NOTE: No DSR processing is yet supported for S3.
    """

    def create_client(self) -> Any:  # type: ignore
        """Returns a client for s3"""
        config = S3Schema(**self.configuration.secrets or {})
        return get_aws_session(
            auth_method=config.auth_method.value,  # pylint: disable=no-member
            storage_secrets=config.model_dump(mode="json"),  # type: ignore
            assume_role_arn=config.aws_assume_role_arn,
        )

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """DSR execution not yet supported for S3"""
        raise NotImplementedError()

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Connects to AWS S3 and gets caller identity to validate credentials.
        """
        logger.info("Starting test connection to {}", self.configuration.key)
        try:
            session = self.client()
            sts_client = session.client("sts")
            sts_client.get_caller_identity()
        except Exception as error:
            raise ConnectionException(str(error))

        return ConnectionTestStatus.succeeded

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """DSR execution not yet supported for S3"""
        return []

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
    ) -> int:
        """DSR execution not yet supported for S3"""
        return 0

    def close(self) -> None:
        """Close any held resources"""
        # no held resources for S3 connector



from typing import Dict, List, Optional
from urllib.parse import urlparse

from fastapi import Body, Depends, HTTPException, Request, Response, Security
from fastapi.responses import PlainTextResponse, RedirectResponse
from fastapi.security import HTTPBasic
from loguru import logger
from sqlalchemy.orm import Session
from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from fides.api.api.deps import get_db
from fides.api.api.v1.endpoints.saas_config_endpoints import (
    verify_oauth_connection_config,
)
from fides.api.common_exceptions import (
    AuthenticationFailure,
    FidesopsException,
    OAuth2TokenException,
)
from fides.api.models.authentication_request import AuthenticationRequest
from fides.api.models.client import ClientDetail
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.models.fides_user import FidesUser
from fides.api.oauth.roles import ROLES_TO_SCOPES_MAPPING
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.client import ClientCreatedResponse
from fides.api.schemas.oauth import AccessToken, OAuth2ClientCredentialsRequestForm
from fides.api.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.service.authentication.authentication_strategy_oauth2_authorization_code import (
    OAuth2AuthorizationCodeAuthenticationStrategy,
)
from fides.api.util.api_router import APIRouter
from fides.api.util.connection_util import connection_status
from fides.common.api.scope_registry import (
    CLIENT_CREATE,
    CLIENT_DELETE,
    CLIENT_READ,
    CLIENT_UPDATE,
    SCOPE_READ,
    SCOPE_REGISTRY,
    ScopeRegistryEnum,
)
from fides.common.api.v1.urn_registry import (
    CLIENT,
    CLIENT_BY_ID,
    CLIENT_SCOPE,
    OAUTH_CALLBACK,
    ROLE,
    SCOPE,
    TOKEN,
    V1_URL_PREFIX,
)
from fides.config import CONFIG

router = APIRouter(tags=["OAuth"], prefix=V1_URL_PREFIX)


@router.post(
    TOKEN,
    response_model=AccessToken,
)
async def acquire_access_token(
    request: Request,
    form_data: OAuth2ClientCredentialsRequestForm = Depends(),
    db: Session = Depends(get_db),
) -> AccessToken:
    """Returns an access token if given credentials are correct, raises 401
    exception if not"""

    basic_credentials = await HTTPBasic(auto_error=False)(request)

    if form_data.client_id and form_data.client_secret:
        client_id = form_data.client_id
        client_secret = form_data.client_secret
    elif basic_credentials:
        client_id = basic_credentials.username
        client_secret = basic_credentials.password
    else:
        raise AuthenticationFailure(detail="Authentication Failure")

    # scopes/roles params are only used if client is root client, otherwise we use the client's associated scopes and/or roles
    client_detail = ClientDetail.get(
        db,
        object_id=client_id,
        config=CONFIG,
        scopes=CONFIG.security.root_user_scopes,
        roles=CONFIG.security.root_user_roles,
    )

    if client_detail is None:
        raise AuthenticationFailure(detail="Authentication Failure")

    if not client_detail.credentials_valid(client_secret):
        raise AuthenticationFailure(detail="Authentication Failure")

    if basic_credentials:
        user = FidesUser.get_by(db, field="username", value=basic_credentials.username)
        if user and user.disabled:  # TODO: Revoke existing session if disabled.
            raise AuthenticationFailure(detail="Authentication Failure")

    logger.info("Creating access token")

    access_code = client_detail.create_access_code_jwe(
        CONFIG.security.app_encryption_key
    )

    if client_id == CONFIG.security.oauth_root_client_id:
        logger.warning(
            "OAuth Root Client ID was used to generate an API access token. If unexpected, review security settings (FIDES__SECURITY__OAUTH_ROOT_CLIENT_ID)"
        )

    return AccessToken(access_token=access_code)


@router.post(
    CLIENT,
    dependencies=[Security(verify_oauth_client, scopes=[CLIENT_CREATE])],
    response_model=ClientCreatedResponse,
)
def create_client(
    *,
    db: Session = Depends(get_db),
    scopes: List[str] = Body([]),
) -> ClientCreatedResponse:
    """Creates a new client and returns the credentials. Only direct scopes can be added to the client via this endpoint."""
    logger.info("Creating new client")
    if not all(scope in SCOPE_REGISTRY for scope in scopes):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Invalid Scope. Scopes must be one of {SCOPE_REGISTRY}.",
        )

    client, secret = ClientDetail.create_client_and_secret(
        db,
        CONFIG.security.oauth_client_id_length_bytes,
        CONFIG.security.oauth_client_secret_length_bytes,
        scopes=scopes,
    )
    return ClientCreatedResponse(client_id=client.id, client_secret=secret)


@router.delete(
    CLIENT_BY_ID, dependencies=[Security(verify_oauth_client, scopes=[CLIENT_DELETE])]
)
def delete_client(client_id: str, db: Session = Depends(get_db)) -> None:
    """Deletes the client associated with the client_id. Does nothing if the client does
    not exist"""
    client = ClientDetail.get(db, object_id=client_id, config=CONFIG)
    if not client:
        return
    logger.info("Deleting client")
    client.delete(db)


@router.get(
    CLIENT_SCOPE,
    dependencies=[Security(verify_oauth_client, scopes=[CLIENT_READ])],
    response_model=List[str],
)
def get_client_scopes(client_id: str, db: Session = Depends(get_db)) -> List[str]:
    """Returns a list of the directly-assigned scopes associated with the client.
    Does not return roles associated with the client.
    Returns an empty list if client does not exist."""
    client = ClientDetail.get(db, object_id=client_id, config=CONFIG)
    if not client:
        return []

    logger.info("Getting client scopes")
    return client.scopes or []


@router.put(
    CLIENT_SCOPE,
    dependencies=[Security(verify_oauth_client, scopes=[CLIENT_UPDATE])],
    response_model=None,
)
def set_client_scopes(
    client_id: str,
    scopes: List[str],
    db: Session = Depends(get_db),
) -> None:
    """Overwrites the client's directly-assigned scopes with those provided.
    Roles cannot be edited via this endpoint.
    Does nothing if the client doesn't exist"""
    client = ClientDetail.get(db, object_id=client_id, config=CONFIG)
    if not client:
        return

    if not all(elem in SCOPE_REGISTRY for elem in scopes):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Invalid Scope. Scopes must be one of {SCOPE_REGISTRY}.",
        )

    logger.info("Updating client scopes")
    client.update(db, data={"scopes": scopes})


@router.get(
    SCOPE,
    dependencies=[Security(verify_oauth_client, scopes=[SCOPE_READ])],
    response_model=List[ScopeRegistryEnum],
)
def read_scopes() -> List[str]:
    """Returns a list of all scopes available for assignment in the system"""
    logger.info("Getting all available scopes")
    return SCOPE_REGISTRY


@router.get(
    ROLE,
    dependencies=[Security(verify_oauth_client, scopes=[SCOPE_READ])],
)
def read_roles_to_scopes_mapping() -> Dict[str, List]:
    """Returns a list of all roles and associated scopes available for assignment in the system"""
    logger.info("Getting all available roles")
    return ROLES_TO_SCOPES_MAPPING


@router.get(OAUTH_CALLBACK)
def oauth_callback(code: str, state: str, db: Session = Depends(get_db)) -> Response:
    """
    Uses the passed in code to generate the token access request
    for the connection associated with the given state.
    """

    # find authentication request by state
    authentication_request: Optional[AuthenticationRequest] = (
        AuthenticationRequest.get_by(db, field="state", value=state)
    )
    if not authentication_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No authentication request found for the given state.",
        )

    connection_config: Optional[ConnectionConfig] = ConnectionConfig.get_by(
        db, field="key", value=authentication_request.connection_key
    )
    verify_oauth_connection_config(connection_config)
    assert connection_config, "Connection config expected!"  # fixes mypy

    try:
        authentication = (
            connection_config.get_saas_config().client_config.authentication  # type: ignore
        )
        auth_strategy: OAuth2AuthorizationCodeAuthenticationStrategy = AuthenticationStrategy.get_strategy(  # type: ignore
            authentication.strategy, authentication.configuration  # type: ignore
        )
        connection_config.secrets = {**connection_config.secrets, "code": code}  # type: ignore
        auth_strategy.get_access_token(connection_config, db)

        msg = f"Test completed for ConnectionConfig with key: {connection_config.key}."
        status_message = connection_status(connection_config, msg, db)

        # default to failed if the status is not set
        test_status_value = (
            status_message.test_status.value
            if status_message.test_status
            else ConnectionTestStatus.failed.value
        )

        if authentication_request.referer:
            # We generate the base URL using only the scheme and netloc (host and port),
            # and omit the path. This is done to ensure the correct redirection URL, as the
            # user might not be on the system's unique page when they follow the
            # authorization link out of Fides.

            parsed_url = urlparse(authentication_request.referer)
            base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"

            if connection_config.system is not None:
                system_key = connection_config.system.fides_key
                redirect_path = f"/systems/configure/{system_key}"
            else:
                redirect_path = f"/integrations/{connection_config.key}"

            return RedirectResponse(
                url=f"{base_url}{redirect_path}?status={test_status_value}"
            )
        return PlainTextResponse(
            content=f"Connection test status: {test_status_value}. No referer URL available. Please navigate back to the Fides Admin UI.",
            status_code=200,
        )
    except (OAuth2TokenException, FidesopsException) as exc:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=str(exc))



"""add asset table

Revision ID: 021166731846
Revises: 58f8edd66b69
Create Date: 2025-01-22 22:14:35.548869

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "021166731846"
down_revision = "58f8edd66b69"
branch_labels = None
depends_on = None


def upgrade():

    op.create_table(
        "asset",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("asset_type", sa.String(), nullable=False),
        sa.Column("domain", sa.String(), nullable=True),
        sa.Column(
            "parent",
            postgresql.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
        sa.Column("parent_domain", sa.String(), nullable=True),
        sa.Column(
            "locations",
            postgresql.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
        sa.Column("with_consent", sa.BOOLEAN(), nullable=False),
        sa.Column(
            "data_uses",
            postgresql.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
        sa.Column(
            "meta",
            postgresql.JSONB(astext_type=sa.Text()),
            server_default="{}",
            nullable=False,
        ),
        sa.Column("base_url", sa.String(), nullable=True),
        sa.Column("system_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(["system_id"], ["ctl_systems.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_asset_asset_type"), "asset", ["asset_type"], unique=False)
    op.create_index(op.f("ix_asset_domain"), "asset", ["domain"], unique=False)
    op.create_index(op.f("ix_asset_id"), "asset", ["id"], unique=False)
    op.create_index(op.f("ix_asset_name"), "asset", ["name"], unique=False)
    op.create_index(op.f("ix_asset_system_id"), "asset", ["system_id"], unique=False)

    op.create_index(
        op.f("ix_asset_name_asset_type_domain_base_url_system_id"),
        "asset",
        [
            "name",
            "asset_type",
            "domain",
            sa.text("coalesce(md5(base_url), 'NULL')"),
            "system_id",
        ],
        unique=True,
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_asset_system_id"), table_name="asset")
    op.drop_index(op.f("ix_asset_name"), table_name="asset")
    op.drop_index(op.f("ix_asset_id"), table_name="asset")
    op.drop_index(op.f("ix_asset_domain"), table_name="asset")
    op.drop_index(op.f("ix_asset_asset_type"), table_name="asset")
    op.drop_index(
        op.f("ix_asset_name_asset_type_domain_base_url_system_id"), table_name="asset"
    )
    op.drop_table("asset")
    # ### end Alembic commands ###



"""update_flexible_legal_basis_default

Revision ID: 1af6950f4625
Revises: 3cc39a6ce32b
Create Date: 2023-11-15 22:15:38.688530

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text
from sqlalchemy.engine import Connection, ResultProxy

revision = "1af6950f4625"
down_revision = "3cc39a6ce32b"
branch_labels = None
depends_on = None


def upgrade():
    bind: Connection = op.get_bind()

    bind.execute(
        text(
            """
            UPDATE privacydeclaration 
            SET flexible_legal_basis_for_processing = false 
            WHERE flexible_legal_basis_for_processing IS NULL;
            """
        )
    )

    op.alter_column(
        "privacydeclaration",
        "flexible_legal_basis_for_processing",
        existing_type=sa.BOOLEAN(),
        server_default="t",
        nullable=False,
    )


def downgrade():
    op.alter_column(
        "privacydeclaration",
        "flexible_legal_basis_for_processing",
        existing_type=sa.BOOLEAN(),
        nullable=True,
        server_default=None,
    )



"""cancel privacy request

Revision ID: ed1b00ff963d
Revises: 55d61eb8ed12
Create Date: 2022-06-03 15:45:14.584540

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ed1b00ff963d"
down_revision = "55d61eb8ed12"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacyrequest", sa.Column("cancel_reason", sa.String(200), nullable=True)
    )
    op.add_column(
        "privacyrequest",
        sa.Column("canceled_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.execute("alter type privacyrequeststatus add value 'canceled'")


def downgrade():
    op.drop_column("privacyrequest", "canceled_at")
    op.drop_column("privacyrequest", "cancel_reason")

    op.execute("delete from privacyrequest where status in ('canceled')")

    op.execute("alter type privacyrequeststatus rename to privacyrequeststatus_old")
    op.execute(
        "create type privacyrequeststatus as enum('in_processing', 'complete', 'pending', 'error', 'paused', 'approved', 'denied')"
    )
    op.execute(
        (
            "alter table privacyrequest alter column status type privacyrequeststatus using "
            "status::text::privacyrequeststatus"
        )
    )
    op.execute("drop type privacyrequeststatus_old")



from typing import Optional

from pydantic import Field, field_validator
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__NOTIFICATIONS__"


class NotificationSettings(FidesSettings):
    """Configuration settings for Data Subject and/or Data Processor notifications."""

    notification_service_type: Optional[str] = Field(
        default=None,
        description="Sets the notification service type used to send notifications. Accepts mailchimp_transactional, mailgun, twilio_sms, or twilio_email.",
    )
    send_request_completion_notification: bool = Field(
        default=False,
        description="When set to True, enables subject notifications upon privacy request completion.",
    )
    send_request_receipt_notification: bool = Field(
        default=False,
        description="When set to True, enables subject notifications upon privacy request receipt.",
    )
    send_request_review_notification: bool = Field(
        default=False,
        description="When set to True, enables subject notifications upon privacy request review.",
    )
    enable_property_specific_messaging: bool = Field(
        default=False,
        description="When set to True, enables property specific messaging feature, otherwise fall back on the messaging template type env flags set above.",
    )

    @field_validator("notification_service_type", mode="before")
    @classmethod
    def validate_notification_service_type(cls, value: Optional[str]) -> Optional[str]:
        """Ensure the provided type is a valid value."""
        if value:
            valid_values = [
                "mailgun",
                "twilio_text",
                "twilio_email",
                "mailchimp_transactional",
                "aws_ses",
            ]
            value = value.lower()  # force lowercase for safety

            if value not in valid_values:
                raise ValueError(
                    f"Invalid NOTIFICATION_SERVICE_TYPE provided '{value}', must be one of: {', '.join([level for level in valid_values])}"
                )

        return value

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



from typing import Optional
from urllib.parse import quote, quote_plus, urlencode

from pydantic import Field, ValidationInfo, field_validator
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__REDIS__"


class RedisSettings(FidesSettings):
    """Configuration settings for Redis."""

    charset: str = Field(
        default="utf8",
        description="Character set to use for Redis, defaults to 'utf8'. Not recommended to change.",
    )
    db_index: int = Field(
        default=0,
        description="The application will use this index in the Redis cache to cache data.",
    )
    test_db_index: int = Field(
        default=1,
        description="The application will use this index in the Redis cache to cache data for testing.",
    )
    decode_responses: bool = Field(
        default=True,
        description="Whether or not to automatically decode the values fetched from Redis. Decodes using the `charset` configuration value.",
    )
    default_ttl_seconds: int = Field(
        default=604800,
        description="The number of seconds for which data will live in Redis before automatically expiring.",
    )
    enabled: bool = Field(
        default=True,
        description="Whether the application's Redis cache should be enabled. Only set to false for certain narrow uses of the application.",
    )
    host: str = Field(
        default="redis",
        description="The network address for the application Redis cache.",
    )
    identity_verification_code_ttl_seconds: int = Field(
        default=600,
        description="Sets TTL for cached identity verification code as part of subject requests.",
    )
    password: str = Field(
        default="testpassword",
        description="The password with which to login to the Redis cache.",
    )
    port: int = Field(
        default=6379,
        description="The port at which the application cache will be accessible.",
    )
    ssl: bool = Field(
        default=False,
        description="Whether the application's connections to the cache should be encrypted using TLS.",
    )
    ssl_cert_reqs: Optional[str] = Field(
        default="required",
        description="If using TLS encryption, set this to 'required' if you wish to enforce the Redis cache to provide a certificate. Note that not all cache providers support this without setting ssl_ca_certs (e.g. AWS Elasticache).",
    )
    ssl_ca_certs: str = Field(
        default="",
        description="If using TLS encryption rooted with a custom Certificate Authority, set this to the path of the CA certificate.",
    )
    user: str = Field(
        default="", description="The user with which to login to the Redis cache."
    )

    # Read-only Redis settings
    read_only_enabled: bool = Field(
        default=False,
        description="Whether a read-only Redis cache is enabled.",
    )
    read_only_host: str = Field(
        default="",
        description="The network address for the read-only Redis cache.",
    )
    # Read-only Redis settings with automatic fallback behavior
    # Field validators will populate these fields with the corresponding writer settings if not provided
    # The order here is important, these must be defined after the writer settings
    read_only_port: int = Field(  # type: ignore[assignment]
        default=None,
        description="The port at which the read-only Redis cache will be accessible. If not provided, the `port` setting will be used.",
    )
    read_only_user: str = Field(  # type: ignore[assignment]
        default=None,
        description="The user with which to login to the read-only Redis cache. If not provided, the `user` setting will be used.",
    )
    read_only_password: str = Field(  # type: ignore[assignment]
        default=None,
        description="The password with which to login to the read-only Redis cache. If not provided, the `password` setting will be used.",
    )
    read_only_db_index: int = Field(  # type: ignore[assignment]
        default=None,
        description="The application will use this index in the read-only Redis cache to cache data. If not provided, the `db_index` setting will be used.",
    )
    read_only_ssl: bool = Field(  # type: ignore[assignment]
        default=None,
        description="Whether the application's connections to the read-only cache should be encrypted using TLS. If not provided, the `ssl` setting will be used.",
    )
    read_only_ssl_cert_reqs: Optional[str] = Field(
        default=None,
        description="If using TLS encryption, set this to 'required' if you wish to enforce the read-only Redis cache to provide a certificate. Note that not all cache providers support this without setting ssl_ca_certs (e.g. AWS Elasticache). If not provided, the `ssl_cert_reqs` setting will be used.",
    )
    read_only_ssl_ca_certs: str = Field(  # type: ignore[assignment]
        default=None,
        description="If using TLS encryption rooted with a custom Certificate Authority, set this to the path of the CA certificate. If not provided, the `ssl_ca_certs` setting will be used.",
    )

    # Field validators for automatic fallback behavior
    @field_validator("read_only_port", mode="before")
    @classmethod
    def resolve_read_only_port(cls, v: Optional[int], info: ValidationInfo) -> int:
        """Resolves the read-only port setting by falling back to the `port` value if no `read_only_port` is specified."""
        if v is None:
            return info.data["port"]
        return v

    @field_validator("read_only_user", mode="before")
    @classmethod
    def resolve_read_only_user(cls, v: Optional[str], info: ValidationInfo) -> str:
        """Resolves the read-only user setting by falling back to the `user` value if no `read_only_user` is specified."""
        if v is None:
            return info.data["user"]
        return v

    @field_validator("read_only_password", mode="before")
    @classmethod
    def resolve_read_only_password(cls, v: Optional[str], info: ValidationInfo) -> str:
        """Resolves the read-only password setting by falling back to the `password` value if no `read_only_password` is specified."""
        if v is None:
            return info.data["password"]
        return v

    @field_validator("read_only_db_index", mode="before")
    @classmethod
    def resolve_read_only_db_index(cls, v: Optional[int], info: ValidationInfo) -> int:
        """Resolves the read-only db_index setting by falling back to the `db_index` value if no `read_only_db_index` is specified."""
        if v is None:
            return info.data["db_index"]
        return v

    @field_validator("read_only_ssl", mode="before")
    @classmethod
    def resolve_read_only_ssl(cls, v: Optional[bool], info: ValidationInfo) -> bool:
        """Resolves the read-only ssl setting by falling back to the `ssl` value if no `read_only_ssl` is specified."""
        if v is None:
            return info.data["ssl"]
        return v

    @field_validator("read_only_ssl_cert_reqs", mode="before")
    @classmethod
    def resolve_read_only_ssl_cert_reqs(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[str]:
        """Resolves the read-only ssl_cert_reqs setting by falling back to the `ssl_cert_reqs` value if no `read_only_ssl_cert_reqs` is specified."""
        if v is None:
            return info.data["ssl_cert_reqs"]
        return v

    @field_validator("read_only_ssl_ca_certs", mode="before")
    @classmethod
    def resolve_read_only_ssl_ca_certs(
        cls, v: Optional[str], info: ValidationInfo
    ) -> str:
        """Resolves the read-only ssl_ca_certs setting by falling back to the `ssl_ca_certs` value if no `read_only_ssl_ca_certs` is specified."""
        if v is None:
            return info.data["ssl_ca_certs"]
        return v

    # This relies on other values to get built so must be last
    connection_url: Optional[str] = Field(
        default=None,
        description="A full connection URL to the Redis cache. If not specified, this URL is automatically assembled from the host, port, password and db_index specified above.",
        exclude=True,
    )
    read_only_connection_url: Optional[str] = Field(
        default=None,
        description="A full connection URL to the read-only Redis cache. If not specified, this URL is automatically assembled from the read_only_host, read_only_port, read_only_password and read_only_db_index specified above.",
        exclude=True,
    )

    @field_validator("connection_url", "read_only_connection_url", mode="before")
    @classmethod
    def assemble_connection_url(
        cls,
        v: Optional[str],
        info: ValidationInfo,
    ) -> str:
        """Join Redis connection credentials into a connection string"""
        if isinstance(v, str):
            # If the whole URL is provided via the config, preference that
            return v

        # Determine which set of settings to use based on field name
        is_read_only = info.field_name == "read_only_connection_url"

        # Extract settings - fallbacks already resolved by field validators for read-only fields
        user = (
            info.data["read_only_user"] if is_read_only else info.data.get("user", "")
        )
        password = (
            info.data["read_only_password"]
            if is_read_only
            else info.data.get("password", "")
        )

        db_index = (
            info.data["read_only_db_index"]
            if is_read_only
            else info.data.get("db_index", "")
        )
        use_tls = (
            info.data["read_only_ssl"] if is_read_only else info.data.get("ssl", False)
        )

        connection_protocol = "redis"
        params_str = ""

        if use_tls:
            # If using TLS update the connection URL format
            connection_protocol = "rediss"

            ssl_cert_reqs = (
                info.data.get("read_only_ssl_cert_reqs", "none")
                if is_read_only
                else info.data.get("ssl_cert_reqs", "none")
            )
            ssl_ca_certs = (
                info.data["read_only_ssl_ca_certs"]
                if is_read_only
                else info.data.get("ssl_ca_certs", "")
            )

            # Build SSL parameters
            params = {"ssl_cert_reqs": quote_plus(ssl_cert_reqs or "none")}
            if ssl_ca_certs:
                params["ssl_ca_certs"] = quote(ssl_ca_certs, safe="/")
            params_str = "?" + urlencode(params, quote_via=quote, safe="/")

        # Configure a basic auth prefix if either user or password is provided, e.g.
        # redis://<user>:<password>@<host>
        auth_prefix = ""
        if password or user:
            auth_prefix = f"{quote_plus(user)}:{quote_plus(password)}@"

        # For host, we don't have a fallback - read replica should be a different host
        host = (
            info.data.get("read_only_host", "")
            if is_read_only
            else info.data.get("host", "")
        )
        port = (
            info.data["read_only_port"] if is_read_only else info.data.get("port", "")
        )
        # Only include database index in URL if it's not the default (0)
        db_path = f"{db_index}" if db_index != 0 else ""

        connection_url = (
            f"{connection_protocol}://{auth_prefix}{host}:{port}/{db_path}{params_str}"
        )
        return connection_url

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



"""add user assigned data_uses and system_id columns to stagedresource

Revision ID: 7c3fbee90c78
Revises: 9a1eacd2666b
Create Date: 2025-03-04 15:00:23.874270

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "7c3fbee90c78"
down_revision = "9a1eacd2666b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "stagedresource",
        sa.Column(
            "user_assigned_system_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "stagedresource",
        sa.Column(
            "user_assigned_data_uses",
            postgresql.ARRAY(sa.VARCHAR()),
            server_default=sa.text("'{}'::character varying[]"),
            autoincrement=False,
            nullable=False,
        ),
    )
    # ### end Alembic commands ###
    op.create_index(
        "ix_stagedresource_user_assigned_system_id",
        "stagedresource",
        ["user_assigned_system_id"],
        unique=False,
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("stagedresource", "user_assigned_data_uses")
    op.drop_column("stagedresource", "user_assigned_system_id")
    # ### end Alembic commands ###
    op.drop_index(
        "ix_stagedresource_user_assigned_system_id", table_name="stagedresource"
    )



from functools import wraps
from typing import Any, Callable, Optional, TypeVar

from loguru import logger
from pydantic import ValidationError

from fides.api.models.manual_tasks.manual_task_log import ManualTaskLog
from fides.api.schemas.manual_tasks.manual_task_config import ManualTaskFieldBase
from fides.api.schemas.manual_tasks.manual_task_schemas import ManualTaskLogStatus
from fides.api.schemas.manual_tasks.manual_task_status import StatusType

T = TypeVar("T")


def validate_fields(fields: list[dict[str, Any]], is_submission: bool = False) -> None:
    """Validate field data against the appropriate Pydantic model.

    Args:
        fields: List of field data to validate
        is_submission: Whether the fields are for a submission (True) or configuration (False)

    Raises:
        ValueError: If the field data is invalid
    """
    # Check for duplicate field keys
    if len({str(field.get("field_key")) for field in fields}) != len(fields):
        raise ValueError(
            "Duplicate field keys found in field updates, field keys must be unique."
        )

    for field in fields:
        if not field.get("field_key"):
            raise ValueError("Invalid field data: field_key is required")
        if not field.get("field_type"):
            raise ValueError("Invalid field data: field_type is required")
        if is_submission and "value" not in field:
            raise ValueError("Invalid field data: value is required for submissions")
        if not is_submission:
            try:
                field_model = ManualTaskFieldBase.get_field_model_for_type(
                    field["field_type"]
                )
                field_model.model_validate(field)
            except ValidationError as e:
                raise ValueError(f"Invalid field data: {str(e)}")


class TaskLogger:
    """Class-based decorator for logging operation success/failure in service methods.
    Args:
        operation_name: The name of the operation to log
        success_status: The status to log for successful operations

    The decorated function can return either:
    - A tuple of (return_value, log_details)
    - Just the return value (in which case log details will be extracted from it)
    """

    def __init__(
        self,
        operation_name: str,
        success_status: ManualTaskLogStatus = ManualTaskLogStatus.complete,
    ) -> None:
        self.operation_name = operation_name
        self.success_status = success_status

    def __call__(self, func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapped(service_self: Any, *args: Any, **kwargs: Any) -> T:
            try:
                result = func(service_self, *args, **kwargs)
                return_value, log_data = (
                    result
                    if isinstance(result, tuple) and len(result) == 2
                    else (result, {})
                )
                if hasattr(service_self, "db") and (
                    task_id := self._get_task_id(return_value, kwargs)
                ):
                    self._log_success(
                        service_self.db, task_id, return_value, log_data, kwargs
                    )
                return return_value
            except Exception as e:
                self._log_error(service_self, e, kwargs)
                raise

        return wrapped

    def _get_task_id(self, result: Any, kwargs: dict) -> Optional[str]:
        """Extract task_id from result or kwargs."""
        return getattr(result, "task_id", None) or kwargs.get("task_id")

    def _get_log_data(self, result: Any, kwargs: dict) -> dict:
        """Create log data from result and kwargs."""
        log_data = {}
        for field in ["task_id", "config_id", "instance_id"]:
            if value := getattr(result, field, None) or kwargs.get(field):
                log_data[field] = value
        log_data["details"] = kwargs.get("details", {})
        return log_data

    def _log_success(
        self, db: Any, task_id: str, result: Any, log_data: dict, kwargs: dict
    ) -> None:
        """Log successful operation."""
        ManualTaskLog.create_log(
            db=db,
            **self._get_log_data(result, kwargs | log_data),
            status=self.success_status,
            message=self.operation_name,
        )

    def _log_error(self, service_self: Any, error: Exception, kwargs: dict) -> None:
        """Log operation error."""
        logger.error(f"Error in {self.operation_name}: {str(error)}")
        if hasattr(service_self, "db") and (task_id := self._get_task_id({}, kwargs)):
            ManualTaskLog.create_log(
                db=service_self.db,
                task_id=task_id,
                **{
                    k: v
                    for k, v in self._get_log_data({}, kwargs).items()
                    if k != "task_id"
                },
                status=ManualTaskLogStatus.error,
                message=f"Error in {self.operation_name}: {str(error)}",
            )

    @staticmethod
    def log_status_change(
        db: Any,
        task_id: str,
        config_id: Optional[str] = None,
        instance_id: Optional[str] = None,
        previous_status: Optional[StatusType] = None,
        new_status: Optional[StatusType] = None,
        user_id: Optional[str] = None,
    ) -> None:
        """Log a status change event."""
        ManualTaskLog.create_log(
            db=db,
            task_id=task_id,
            config_id=config_id,
            instance_id=instance_id,
            status=ManualTaskLogStatus.in_progress,
            message=f"Task instance status transitioning from {previous_status} to {new_status}",
            details={
                "previous_status": previous_status,
                "new_status": new_status,
                "user_id": user_id,
            },
        )

    @staticmethod
    def log_create(
        db: Any,
        task_id: str,
        entity_type: str,
        entity_id: str,
        details: Optional[dict] = None,
        user_id: Optional[str] = None,
        status: ManualTaskLogStatus = ManualTaskLogStatus.created,
    ) -> None:
        """Log a creation event."""
        id_fields = {
            "task": {"task_id": task_id},
            "config": {"task_id": task_id, "config_id": entity_id},
            "instance": {"task_id": task_id, "instance_id": entity_id},
            "submission": {"task_id": task_id, "instance_id": entity_id},
        }
        if entity_type not in id_fields:
            raise ValueError(f"Invalid entity type: {entity_type}")

        ManualTaskLog.create_log(
            db=db,
            **id_fields[entity_type],
            status=status,
            message=f"Created new {entity_type}",
            details=(details or {}) | ({"user_id": user_id} if user_id else {}),
        )


# Alias the class to maintain the original decorator name
with_task_logging = TaskLogger



from enum import Enum
from io import BufferedReader
from typing import Any, Dict, List, Optional, Tuple

from pydantic import BaseModel, ConfigDict

RequestFile = Tuple[str, Tuple[str, BufferedReader, str]]


class HTTPMethod(Enum):
    """Enum to represent HTTP Methods"""

    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"
    HEAD = "HEAD"


class SaaSRequestParams(BaseModel):
    """
    Holds the method, path, headers, query, and body params to build a SaaS HTTP request.
    """

    method: HTTPMethod
    path: str
    headers: Dict[str, Any] = {}
    query_params: Dict[str, Any] = {}
    body: Optional[str] = None
    files: Optional[List[RequestFile]] = None
    model_config = ConfigDict(use_enum_values=True, arbitrary_types_allowed=True)


class ConnectorParamRef(BaseModel):
    """A reference to a value in the connector params (by name)"""

    connector_param: Any = None


class IdentityParamRef(BaseModel):
    """A reference to the identity type in the filter Post Processor Config"""

    identity: str


class DatasetRef(BaseModel):
    """A reference to the dataset field in the filter Post Processor Config"""

    dataset_reference: str


class ConsentPropagationStatus(Enum):
    """
    An enum for the different statuses that can be returned from a consent propagation request.
    """

    executed = "executed"
    no_update_needed = "no_update_needed"
    missing_data = "missing_data"



"""Various functions to help serve the UI from our server"""

import importlib.util
import re
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional, Union

from fastapi import Response
from fastapi.responses import FileResponse
from loguru import logger

FIDES_DIRECTORY = "src/fides"
ADMIN_UI_DIRECTORY = "ui-build/static/admin/"


def get_package_path() -> Optional[Path]:
    """
    Returns a Path to the root directory of this package's installation, if it exists.

    If installed via `pip install -e` (like in Dockerfile's `dev` target), this may be /fides/src/fides
    If installed via `pip install ethyca-fides`, this will be more like /usr/local/lib/python/site-packages/fides
    """
    package_name = __package__.split(".", maxsplit=1)[0]
    spec = importlib.util.find_spec(package_name)
    if spec and spec.origin:
        return Path(spec.origin).parent

    return None


def get_path_to_admin_ui_file(path: str = "") -> Optional[Path]:
    """
    Return a path to a packaged admin UI file.

    If no path is given, returns the path to the root of the UI directory
    """
    package_path = get_package_path()
    if package_path is None:
        return None

    return package_path / ADMIN_UI_DIRECTORY / path


def get_admin_index_as_response() -> Response:
    """Get just the frontend index file as a FileResponse.
    If it does not exist, return a placeholder."""
    placeholder = "<h1>Privacy is a Human Right!</h1>"
    index = get_path_to_admin_ui_file("index.html")
    return FileResponse(index) if index and index.is_file() else Response(placeholder)


@lru_cache
def get_ui_file_map() -> Dict[re.Pattern, Path]:
    """Get the Admin UI route map from its static files."""
    package_path = get_package_path()
    if package_path is None:
        return {}

    return generate_route_file_map(package_path / ADMIN_UI_DIRECTORY)


def generate_route_file_map(ui_directory: Union[str, Path]) -> Dict[re.Pattern, Path]:
    """
    Generates a map of route patterns to the paths of files that route should serve.
    The returned paths are absolute (not relative to ui_directory or CWD).

    :param ui_directory: The path (or str) of the directory to search for route-able files.
    """
    ui_path = Path(ui_directory)
    if not ui_path.exists():
        return {}

    exact_pattern = r"\[[a-zA-Z.[\]]+\]"
    nested_pattern = r"\[...[a-zA-Z.[\]]+\]"

    exact_pattern_replacement = r"[a-zA-Z10-9-_.[\]]+"
    nested_pattern_replacement = r"[a-zA-Z10-9-_.[\]/]+"

    route_file_map = {}

    for filepath in ui_path.glob("**/*.html"):
        # Ignore directory root.
        if filepath == ui_path:
            continue

        # Path within the ui_directory, with the file extension (.html) removed.
        relative_path = filepath.relative_to(ui_path).with_suffix("")

        route = str(relative_path)
        if re.search(exact_pattern, str(relative_path)):
            route = re.sub(exact_pattern, exact_pattern_replacement, str(relative_path))
        if re.search(nested_pattern, str(filepath)):
            route = re.sub(
                nested_pattern,
                nested_pattern_replacement,
                str(relative_path),
            )

        # Full match, allowing trailing slash.
        pattern = re.compile(r"^" + route + r"/?$")

        route_file_map[pattern] = filepath

    return route_file_map


def match_route(route_file_map: Dict[re.Pattern, Path], route: str) -> Optional[Path]:
    """
    Match a route against a route file map and return the match, if any.

    If multiple routes match, prefers static routes (e.g. dataset/new) over
    dynamic (e.g. dataset/[id]).
    """
    matches = [
        path for pattern, path in route_file_map.items() if re.fullmatch(pattern, route)
    ]

    if not matches:
        return None

    if len(matches) == 1:
        return matches[0]

    # Multiple matches were found, so look for the first static path
    static_matches = [path for path in matches if not _is_dynamic_path(path)]
    if static_matches:
        return sorted(static_matches)[0]

    # If no static matches exist, fallback to the first match
    return sorted(matches)[0]


def path_is_in_ui_directory(path: Path) -> bool:
    """Checks if the path exists within the UI directory"""
    ui_directory = get_path_to_admin_ui_file()
    if not ui_directory:
        logger.debug("Unable to locate UI directory")
        return False

    return ui_directory in path.parents


def _is_dynamic_path(path: Path) -> bool:
    """Returns true if the given route is a dynamic NextJS route (e.g. "dataset/[id].html")"""
    return re.compile(r"\[\w+\]").search(str(path)) is not None



from typing import ClassVar, List

from pydantic import Field

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class OktaSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to Okta"""

    org_url: str = Field(
        title="Organization URL",
        description="The URL of your Okta organization (e.g. https://your-org.okta.com)",
    )
    api_token: str = Field(
        title="API Token",
        description="The API token used to authenticate with Okta",
        json_schema_extra={"sensitive": True},
    )

    _required_components: ClassVar[List[str]] = ["org_url", "api_token"]


class OktaDocsSchema(OktaSchema, NoValidationSchema):
    """Okta Secrets Schema for API Docs"""



"""
AES GCM encryption utilities with SQLAlchemy-Utils and cryptography library implementations.

This module provides simplified encrypt/decrypt functions using two approaches:
1. SQLAlchemy-Utils AesGcmEngine (compatible with existing database encryption)
2. Cryptography library with chunked processing (better performance, standard library)
"""

import base64
import hashlib
import json
import os
from typing import Any, List, Optional, Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from loguru import logger
from sqlalchemy_utils.types.encrypted.encrypted_type import AesGcmEngine

from fides.api.util.collection_util import Row
from fides.api.util.custom_json_encoder import CustomJSONEncoder, _custom_decoder
from fides.config import CONFIG


class EncryptionError(Exception):
    """Raised when encryption/decryption operations fail"""


# SQLAlchemy-Utils Implementation (for compatibility with existing database encryption)
def encrypt_with_sqlalchemy_utils(data: List[Row]) -> bytes:
    """
    Serialize and encrypt data using CustomJSONEncoder and SQLAlchemy-Utils AesGcmEngine.

    This approach is compatible with existing database encryption but has lower performance.

    Args:
        data: Raw data to serialize and encrypt

    Returns:
        Encrypted bytes

    Raises:
        EncryptionError: If serialization or encryption fails
    """
    try:
        # Serialize using CustomJSONEncoder for consistent ObjectId handling
        serialized_data = json.dumps(data, cls=CustomJSONEncoder, separators=(",", ":"))
        data_bytes = serialized_data.encode("utf-8")

        # Encrypt using SQLAlchemy-Utils AesGcmEngine
        engine = AesGcmEngine()
        key = CONFIG.security.app_encryption_key
        engine._update_key(key)  # pylint: disable=protected-access

        # AesGcmEngine expects string input
        data_str = data_bytes.decode("utf-8")
        encrypted_data = engine.encrypt(data_str)
        encrypted_bytes = encrypted_data.encode("utf-8")

        logger.debug(
            f"SQLAlchemy-Utils: Encrypted {len(data_bytes)} bytes to {len(encrypted_bytes)} bytes"
        )
        return encrypted_bytes

    except Exception as e:
        logger.error(f"SQLAlchemy-Utils encryption failed: {e}")
        raise EncryptionError(f"SQLAlchemy-Utils encryption failed: {str(e)}")


def decrypt_with_sqlalchemy_utils(encrypted_bytes: bytes) -> List[Row]:
    """
    Decrypt and deserialize data using SQLAlchemy-Utils AesGcmEngine and _custom_decoder.

    Args:
        encrypted_bytes: Encrypted data bytes to decrypt

    Returns:
        Deserialized data

    Raises:
        EncryptionError: If decryption or deserialization fails
    """
    try:
        # Decrypt using SQLAlchemy-Utils AesGcmEngine
        engine = AesGcmEngine()
        key = CONFIG.security.app_encryption_key
        engine._update_key(key)  # pylint: disable=protected-access

        # AesGcmEngine expects string input
        encrypted_str = encrypted_bytes.decode("utf-8")
        decrypted_data = engine.decrypt(encrypted_str)

        # Deserialize using _custom_decoder for consistent ObjectId handling
        data = json.loads(decrypted_data, object_hook=_custom_decoder)

        logger.debug(
            f"SQLAlchemy-Utils: Decrypted {len(encrypted_bytes)} bytes to {len(data)} records"
        )
        return data

    except Exception as e:
        logger.error(f"SQLAlchemy-Utils decryption failed: {e}")
        raise EncryptionError(f"SQLAlchemy-Utils decryption failed: {str(e)}")


# Cryptography Library Implementation (standard, chunked processing)
def encrypt_with_cryptography(
    data: Union[List[Row], Any], chunk_size: Optional[int] = None
) -> bytes:
    """
    Serialize and encrypt data using the standard cryptography library with chunked processing.

    This provides fast performance and memory efficiency for large datasets.

    Args:
        data: Raw data to serialize and encrypt
        chunk_size: Size of chunks for processing (default 4MB)

    Returns:
        Encrypted bytes (base64-encoded string as bytes)

    Raises:
        EncryptionError: If serialization or encryption fails
    """
    try:
        # Set default chunk size
        if chunk_size is None:
            chunk_size = 4 * 1024 * 1024  # 4MB chunks

        # Serialize using CustomJSONEncoder for consistent handling
        serialized_data = json.dumps(data, cls=CustomJSONEncoder, separators=(",", ":"))
        plaintext = serialized_data.encode("utf-8")

        data_size_mb = len(plaintext) / (1024 * 1024)
        chunk_size_mb = chunk_size / (1024 * 1024)
        estimated_chunks = len(plaintext) // chunk_size + (
            1 if len(plaintext) % chunk_size else 0
        )
        record_count = len(data) if isinstance(data, list) else "N/A"

        logger.info(
            f"Cryptography: Encrypting {record_count} records ({data_size_mb:.1f} MB) "
            f"using {chunk_size_mb:.0f}MB chunks (~{estimated_chunks} chunks)"
        )

        # Use SQLAlchemy-Utils compatible key (SHA256 hash of app key)
        key = _get_sqlalchemy_compatible_key()
        nonce = os.urandom(12)  # 96-bit nonce for AES-GCM

        # Create cipher
        cipher = Cipher(
            algorithms.AES(key), modes.GCM(nonce), backend=default_backend()
        )
        encryptor = cipher.encryptor()

        # Process in chunks for memory efficiency
        ciphertext_chunks = []
        for i in range(0, len(plaintext), chunk_size):
            chunk = plaintext[i : i + chunk_size]
            ciphertext_chunks.append(encryptor.update(chunk))

        # Finalize and get tag
        encryptor.finalize()
        tag = encryptor.tag

        # Combine in same format as SQLAlchemy-Utils: [nonce/iv][tag][ciphertext]
        ciphertext = b"".join(ciphertext_chunks)
        binary_result = nonce + tag + ciphertext

        # Base64 encode to match SQLAlchemy-Utils format
        base64_result = base64.b64encode(binary_result).decode("utf-8")
        result_bytes = base64_result.encode("utf-8")

        encrypted_size_mb = len(result_bytes) / (1024 * 1024)
        logger.info(
            f"Cryptography: Encrypted successfully - "
            f"{len(ciphertext_chunks)} chunks, {encrypted_size_mb:.1f} MB output (base64)"
        )

        return result_bytes

    except Exception as e:
        logger.error(f"Cryptography encryption failed: {e}")
        raise EncryptionError(f"Cryptography encryption failed: {str(e)}")


def decrypt_with_cryptography(
    encrypted_bytes: bytes, chunk_size: Optional[int] = None
) -> Union[List[Row], Any]:
    """
    Decrypt and deserialize data using the cryptography library with chunked processing.

    Args:
        encrypted_bytes: Encrypted data (base64-encoded string as bytes)
        chunk_size: Size of chunks for processing (default 4MB)

    Returns:
        Deserialized data

    Raises:
        EncryptionError: If decryption or deserialization fails
    """
    try:
        # Set default chunk size
        if chunk_size is None:
            chunk_size = 4 * 1024 * 1024  # 4MB chunks

        # Decode from base64
        encrypted_str = encrypted_bytes.decode("utf-8")
        binary_data = base64.b64decode(encrypted_str)

        # Extract components in SQLAlchemy-Utils format: [nonce/iv][tag][ciphertext]
        if len(binary_data) < 28:  # 12 (nonce) + 16 (tag)
            raise ValueError("Encrypted data too short")

        nonce = binary_data[:12]  # First 12 bytes: nonce/IV
        tag = binary_data[12:28]  # Next 16 bytes: tag
        ciphertext = binary_data[28:]  # Remaining bytes: ciphertext

        encrypted_size_mb = len(encrypted_bytes) / (1024 * 1024)
        chunk_size_mb = chunk_size / (1024 * 1024)
        estimated_chunks = len(ciphertext) // chunk_size + (
            1 if len(ciphertext) % chunk_size else 0
        )

        logger.info(
            f"Cryptography: Decrypting {encrypted_size_mb:.1f} MB "
            f"using {chunk_size_mb:.0f}MB chunks (~{estimated_chunks} chunks)"
        )

        # Use SQLAlchemy-Utils compatible key
        key = _get_sqlalchemy_compatible_key()
        cipher = Cipher(
            algorithms.AES(key), modes.GCM(nonce, tag), backend=default_backend()
        )
        decryptor = cipher.decryptor()

        # Process in chunks for memory efficiency
        plaintext_chunks = []
        for i in range(0, len(ciphertext), chunk_size):
            chunk = ciphertext[i : i + chunk_size]
            plaintext_chunks.append(decryptor.update(chunk))

        # Finalize
        decryptor.finalize()

        # Combine and deserialize
        plaintext = b"".join(plaintext_chunks)
        decrypted_json = plaintext.decode("utf-8")
        data = json.loads(decrypted_json, object_hook=_custom_decoder)

        record_count = len(data) if isinstance(data, list) else "N/A"
        logger.info(f"Cryptography: Successfully decrypted {record_count} records")

        return data

    except Exception as e:
        logger.error(f"Cryptography decryption failed: {e}")
        raise EncryptionError(f"Cryptography decryption failed: {str(e)}")


def _get_sqlalchemy_compatible_key() -> bytes:
    """Get 32-byte encryption key compatible with SQLAlchemy-Utils AesGcmEngine."""
    app_key = CONFIG.security.app_encryption_key.encode(CONFIG.security.encoding)
    # SQLAlchemy-Utils always uses SHA256 hash of the key
    return hashlib.sha256(app_key).digest()


# Public API - Use cryptography by default for new operations
encrypt_data = encrypt_with_cryptography
decrypt_data = decrypt_with_cryptography



from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal, Optional

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import FidesopsException, OAuth2TokenException
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.schemas.saas.saas_config import ClientConfig, SaaSRequest
from fides.api.schemas.saas.strategy_configuration import OAuth2BaseConfiguration
from fides.api.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.service.connectors.saas.authenticated_client import AuthenticatedClient
from fides.api.util.logger import Pii
from fides.api.util.saas_util import assign_placeholders, map_param_values


class OAuth2AuthenticationStrategyBase(AuthenticationStrategy):
    """
    Base class for the various OAuth2 flows. Includes common functionality
    for processing token requests and responses.
    """

    def __init__(self, configuration: OAuth2BaseConfiguration) -> None:
        self.expires_in = configuration.expires_in
        self.token_request = configuration.token_request
        self.refresh_request = configuration.refresh_request

    @property
    def _required_secrets(self) -> List[str]:
        """A list of required secrets for the given OAuth2 strategy."""
        return ["client_id", "client_secret"]

    @staticmethod
    def _close_to_expiration(
        expires_at: int, connection_config: ConnectionConfig
    ) -> bool:
        """Check if the access_token will expire in the next 10 minutes."""

        if expires_at is None:
            logger.info(
                "The expires_at value is not defined for {}, skipping token refresh",
                connection_config.key,
            )
            return False

        return expires_at < (datetime.utcnow() + timedelta(minutes=10)).timestamp()

    @staticmethod
    def _call_token_request(
        action: Literal["access", "refresh"],
        token_request: SaaSRequest,
        connection_config: ConnectionConfig,
    ) -> Dict[str, Any]:
        """
        Generates and executes the token request based on the OAuth2 config
        and connection config secrets.
        """

        logger.info("Attempting {} token request for {}", action, connection_config.key)

        # get the client config from the token request or default to the
        # protocol and host specified by the root client config (no auth)
        root_client_config = connection_config.get_saas_config().client_config  # type: ignore
        oauth_client_config = (
            token_request.client_config
            if token_request.client_config
            else ClientConfig(
                protocol=root_client_config.protocol, host=root_client_config.host
            )
        )

        client = AuthenticatedClient(
            (
                f"{oauth_client_config.protocol}://"
                f"{assign_placeholders(oauth_client_config.host, connection_config.secrets)}"  # type: ignore
            ),
            connection_config,
            oauth_client_config,
        )

        try:
            # map param values to placeholders in request
            prepared_request = map_param_values(
                action,
                f"{connection_config.name} OAuth2",
                token_request,
                connection_config.secrets,  # type: ignore
            )
            # ignore errors so we can return the error message in the response
            response = client.send(prepared_request, ignore_errors=True)
            json_response = response.json()
        except Exception as exc:
            logger.error(
                "Error occurred during the {} request for {}: {}",
                action,
                connection_config.key,
                Pii(str(exc)),
            )
            raise OAuth2TokenException(
                f"Error occurred during the {action} request for {connection_config.key}: {str(exc)}"
            )

        return json_response

    def _check_required_secrets(self, connection_config: ConnectionConfig) -> None:
        """
        Verifies that the secrets required for authentication have been provided.
        """

        secrets = connection_config.secrets or {}
        required_secrets = {name: secrets.get(name) for name in self._required_secrets}
        if not all(required_secrets.values()):
            missing_secrets = [
                name for name, value in required_secrets.items() if not value
            ]
            raise FidesopsException(
                f"Missing required secret(s) '{', '.join(missing_secrets)}' for {connection_config.key}"
            )

    def _validate_and_store_response(
        self,
        response: Dict[str, Any],
        connection_config: ConnectionConfig,
        db: Optional[Session] = None,
    ) -> str:
        """
        Persists and returns the new access token.
        Also updates the refresh token if one is provided.
        """

        access_token = response.get("access_token")

        # error, error_description, and error_uri are part of the OAuth2 spec
        if access_token is None:
            error_message = " ".join(
                filter(
                    None,
                    (
                        f"Unable to retrieve token for {connection_config.key} ({response.get('error')}).",
                        response.get("error_description"),
                        response.get("error_uri"),
                    ),
                )
            )
            logger.error(error_message)
            raise OAuth2TokenException(error_message)

        data = {"access_token": access_token}

        # The authorization server MAY issue a new refresh token, in which case
        # the client MUST discard the old refresh token and replace it with the
        # new refresh token.
        #
        # https://datatracker.ietf.org/doc/html/rfc6749#section-6

        refresh_token = response.get("refresh_token")
        if refresh_token:
            data["refresh_token"] = refresh_token

        # If omitted, the authorization server SHOULD provide the
        # expiration time via other means or document the default value.
        #
        # https://datatracker.ietf.org/doc/html/rfc6749#section-5.1
        #
        # This alternate way of specifying the expiration is handled
        # by the optional expires_in field of the OAuth2AuthenticationConfiguration

        expires_in = self.expires_in or response.get("expires_in")
        if expires_in:
            data["expires_at"] = int(datetime.utcnow().timestamp()) + expires_in

        # persist new tokens to the database
        # ideally we use a passed in database session but we can
        # get the session from the connection_config as a fallback
        db = db or Session.object_session(connection_config)
        updated_secrets = {**connection_config.secrets, **data}  # type: ignore
        connection_config.update(db, data={"secrets": updated_secrets})
        logger.info(
            "Successfully updated the OAuth2 token(s) for {}", connection_config.key
        )

        return access_token

    def get_access_token(
        self, connection_config: ConnectionConfig, db: Optional[Session] = None
    ) -> str:
        """
        Executes the access token request based on the OAuth2 config
        and connection config secrets. The access and refresh tokens returned from
        this request are stored in the connection config secrets.
        """

        self._check_required_secrets(connection_config)
        access_response = self._call_token_request(
            "access", self.token_request, connection_config
        )
        return self._validate_and_store_response(access_response, connection_config, db)

    def _refresh_token(self, connection_config: ConnectionConfig) -> str:
        """
        Persists and returns a refreshed access_token if the token is close to expiring.
        Otherwise just returns the existing access_token.
        """

        if self.refresh_request:
            expires_at = connection_config.secrets.get("expires_at")  # type: ignore
            if self._close_to_expiration(expires_at, connection_config):
                refresh_response = self._call_token_request(
                    "refresh", self.refresh_request, connection_config
                )
                return self._validate_and_store_response(
                    refresh_response, connection_config
                )
        return connection_config.secrets.get("access_token")  # type: ignore



"""add privacynotice tables

Revision ID: 6d6b0b7cbb36
Revises: 8615ac51e78b
Create Date: 2023-03-22 12:09:47.395190

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6d6b0b7cbb36"
down_revision = "8615ac51e78b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "privacynotice",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("origin", sa.String(), nullable=True),
        sa.Column("regions", sa.ARRAY(sa.String()), nullable=False),
        sa.Column(
            "consent_mechanism",
            sa.Enum("opt_in", "opt_out", "necessary", name="consentmechanism"),
            nullable=False,
        ),
        sa.Column("data_uses", sa.ARRAY(sa.String()), nullable=False),
        sa.Column("version", sa.Float(), nullable=False),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_privacy_center", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_banner", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_privacy_modal", sa.Boolean(), nullable=False),
        sa.Column("enforcement_level", sa.String(), nullable=False),
        sa.Column("has_gpc_flag", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_privacynotice_id"), "privacynotice", ["id"], unique=False)
    op.create_index(
        op.f("ix_privacynotice_regions"), "privacynotice", ["regions"], unique=False
    )
    op.create_table(
        "privacynoticehistory",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("origin", sa.String(), nullable=True),
        sa.Column("regions", sa.ARRAY(sa.String()), nullable=False),
        sa.Column(
            "consent_mechanism",
            sa.Enum("opt_in", "opt_out", "necessary", name="consentmechanism"),
            nullable=False,
        ),
        sa.Column("data_uses", sa.ARRAY(sa.String()), nullable=False),
        sa.Column("version", sa.Float(), nullable=False),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_privacy_center", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_banner", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_privacy_modal", sa.Boolean(), nullable=False),
        sa.Column("enforcement_level", sa.String(), nullable=False),
        sa.Column("has_gpc_flag", sa.Boolean(), nullable=False),
        sa.Column("privacy_notice_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_notice_id"],
            ["privacynotice.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacynoticehistory_id"), "privacynoticehistory", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_privacynoticehistory_regions"),
        "privacynoticehistory",
        ["regions"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_privacynoticehistory_regions"), table_name="privacynoticehistory"
    )
    op.drop_index(op.f("ix_privacynoticehistory_id"), table_name="privacynoticehistory")
    op.drop_table("privacynoticehistory")
    op.drop_index(op.f("ix_privacynotice_regions"), table_name="privacynotice")
    op.drop_index(op.f("ix_privacynotice_id"), table_name="privacynotice")
    op.drop_table("privacynotice")
    # ### end Alembic commands ###

    # specifically drop the enum type since it's not being deleted during table drop
    op.execute("""DROP TYPE consentmechanism""")



from datetime import datetime
from typing import Any, Optional, Type

from loguru import logger

from fides.api.api.deps import get_autoclose_db_session
from fides.api.schemas.external_storage import ExternalStorageMetadata
from fides.api.service.external_data_storage import (
    ExternalDataStorageError,
    ExternalDataStorageService,
)
from fides.api.util.data_size import LARGE_DATA_THRESHOLD_BYTES, calculate_data_size


class EncryptedLargeDataDescriptor:
    """
    A Python descriptor for database fields with encrypted external storage fallback.

    See the original implementation for detailed docstrings.
    """

    def __init__(
        self,
        field_name: str,
        empty_default: Optional[Any] = None,
        threshold_bytes: Optional[int] = None,
    ):
        self.field_name = field_name
        self.private_field = f"_{field_name}"
        self.empty_default = empty_default if empty_default is not None else []
        self.threshold_bytes = threshold_bytes or LARGE_DATA_THRESHOLD_BYTES
        self.model_class: Optional[str] = None
        self.name: Optional[str] = None

    # Descriptor protocol helpers

    def __set_name__(
        self, owner: Type, name: str
    ) -> None:  # noqa: D401 (docstring in orig file)
        self.name = name
        self.model_class = owner.__name__

    def _generate_storage_path(self, instance: Any) -> str:
        instance_id = getattr(instance, "id", None)
        if not instance_id:
            raise ValueError(f"Instance {instance} must have an 'id' attribute")
        timestamp = datetime.utcnow().strftime("%Y%m%d-%H%M%S-%f")
        return f"{self.model_class}/{instance_id}/{self.field_name}/{timestamp}.txt"

    def __get__(self, instance: Any, owner: Type) -> Any:  # noqa: D401
        if instance is None:
            return self
        raw_data = getattr(instance, self.private_field)
        if raw_data is None:
            return None
        if isinstance(raw_data, dict) and "storage_type" in raw_data:
            logger.info(
                f"Reading {self.model_class}.{self.field_name} from external storage "
                f"({raw_data.get('storage_type')})"
            )
            try:
                metadata = ExternalStorageMetadata.model_validate(raw_data)
                data = self._retrieve_external_data(metadata)
                record_count = len(data) if isinstance(data, list) else "N/A"
                logger.info(
                    f"Successfully retrieved {self.model_class}.{self.field_name} "
                    f"from external storage (records: {record_count})"
                )
                return data if data is not None else self.empty_default
            except Exception as e:  # pylint: disable=broad-except
                logger.error(
                    f"Failed to retrieve {self.model_class}.{self.field_name} "
                    f"from external storage: {str(e)}"
                )
                raise ExternalDataStorageError(
                    f"Failed to retrieve {self.field_name}: {str(e)}"
                ) from e
        else:
            return raw_data

    def __set__(self, instance: Any, value: Any) -> None:  # noqa: D401
        if not value:
            self._cleanup_external_data(instance)
            setattr(instance, self.private_field, self.empty_default)
            return
        try:
            current_data = self.__get__(instance, type(instance))
            if current_data == value:
                return
        except Exception:  # pylint: disable=broad-except
            pass

        data_size = calculate_data_size(value)
        if data_size > self.threshold_bytes:
            logger.info(
                f"{self.model_class}.{self.field_name}: Data size ({data_size:,} bytes) "
                f"exceeds threshold ({self.threshold_bytes:,} bytes), storing externally"
            )
            self._cleanup_external_data(instance)
            metadata = self._store_external_data(instance, value)
            setattr(instance, self.private_field, metadata.model_dump())
        else:
            self._cleanup_external_data(instance)
            setattr(instance, self.private_field, value)

    # External storage helpers

    def _store_external_data(self, instance: Any, data: Any) -> ExternalStorageMetadata:
        storage_path = self._generate_storage_path(instance)
        with get_autoclose_db_session() as session:
            metadata = ExternalDataStorageService.store_data(
                db=session,
                storage_path=storage_path,
                data=data,
            )
            logger.info(
                f"Stored {self.model_class}.{self.field_name} to external storage: {storage_path}"
            )
            return metadata

    @staticmethod
    def _retrieve_external_data(metadata: ExternalStorageMetadata) -> Any:  # noqa: D401
        with get_autoclose_db_session() as session:
            return ExternalDataStorageService.retrieve_data(
                db=session,
                metadata=metadata,
            )

    def _cleanup_external_data(self, instance: Any) -> None:  # noqa: D401
        raw_data = getattr(instance, self.private_field, None)
        if isinstance(raw_data, dict) and "storage_type" in raw_data:
            try:
                metadata = ExternalStorageMetadata.model_validate(raw_data)
                with get_autoclose_db_session() as session:
                    ExternalDataStorageService.delete_data(
                        db=session,
                        metadata=metadata,
                    )
                logger.info(
                    f"Cleaned up external storage for {self.model_class}.{self.field_name}: "
                    f"{metadata.file_key}"
                )
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(
                    f"Failed to cleanup external {self.field_name}: {str(e)}"
                )

    # Public helper

    def cleanup(self, instance: Any) -> None:  # noqa: D401
        self._cleanup_external_data(instance)



from enum import Enum
from typing import Optional


class TestStatus(Enum):
    passed = "passed"
    failed = "failed"
    untested = "untested"

    def str_to_bool(self) -> Optional[bool]:
        """Translates query param string to optional/bool value
        for filtering ConnectionConfig.last_test_succeeded field"""
        if self == self.passed:
            return True
        if self == self.failed:
            return False
        return None



"""Merge heads

Revision ID: 4fc34906c389
Revises: aaa81d97a6f6, f131a1263a10
Create Date: 2022-08-17 19:35:33.413593

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4fc34906c389"
down_revision = ("aaa81d97a6f6", "f131a1263a10")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""Add azure as a SSO provider

Revision ID: 1c8dfd6a1bc6
Revises: 4ebe0766021b
Create Date: 2024-10-21 18:04:29.130138

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "1c8dfd6a1bc6"
down_revision = "4ebe0766021b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("alter type providerenum add value 'azure'")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("delete from openid_provider where provider in ('azure')")
    op.execute("alter type providerenum rename to providerenum_old")
    op.execute("create type providerenum as enum('google','okta','custom')")
    op.execute(
        (
            "alter table openid_provider alter column provider type providerenum using "
            "openid_provider::text::providerenum;"
        )
    )
    op.execute("drop type providerenum_old")
    # ### end Alembic commands ###



"""adds fides_user_device_id enum

Revision ID: 3842d1acac5f
Revises: 8342453518cc
Create Date: 2023-04-21 20:10:01.389078

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "3842d1acac5f"
down_revision = "8342453518cc"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("ALTER TYPE providedidentitytype RENAME TO providedidentitytype_old")
    op.execute(
        "CREATE TYPE providedidentitytype AS ENUM('email', 'phone_number', 'ga_client_id', 'ljt_readerID', 'fides_user_device_id')"
    )
    op.execute(
        "ALTER TABLE providedidentity ALTER COLUMN field_name TYPE providedidentitytype USING field_name::text::providedidentitytype"
    )
    op.execute("DROP TYPE providedidentitytype_old")


def downgrade():
    op.execute("ALTER TYPE providedidentitytype RENAME TO providedidentitytype_old")
    op.execute(
        "CREATE TYPE providedidentitytype AS ENUM('email', 'phone_number', 'ga_client_id', 'ljt_readerID')"
    )
    op.execute(
        "ALTER TABLE providedidentity ALTER COLUMN field_name TYPE providedidentitytype USING field_name::text::providedidentitytype"
    )
    op.execute("DROP TYPE providedidentitytype_old")



"""data migration converting existing 'Cookies' resources to 'Asset' resources with type 'Cookie'

Revision ID: c9c72b3d550b
Revises: 9288f729cac4
Create Date: 2025-02-18 18:33:56.039924

"""

import uuid
from typing import Dict, Tuple

import sqlalchemy as sa
from alembic import op
from loguru import logger

from fides.api.models.asset import Asset

# revision identifiers, used by Alembic.
revision = "c9c72b3d550b"
down_revision = "9288f729cac4"
branch_labels = None
depends_on = None


def upgrade():
    # migrate existing cookies to assets
    connection = op.get_bind()
    result = connection.execute(
        sa.text(
            "SELECT c.id, c.created_at, c.updated_at, c.name, c.domain, c.system_id, c.path, c.privacy_declaration_id, pd.system_id AS pud_system_id, pd.data_use "
            "FROM cookies c "
            "LEFT JOIN privacydeclaration pd ON c.privacy_declaration_id = pd.id"
        )
    )

    logger.debug("Converting existing cookies to assets")
    assets_to_create: Dict[str, Asset] = {}
    for row in result:
        if row.privacy_declaration_id is None:
            identifier = f"{row.name}_{row.system_id}"
            if identifier not in assets_to_create.keys():
                assets_to_create[identifier] = Asset(
                    id=str(uuid.uuid4()),
                    created_at=row.created_at,
                    updated_at=row.updated_at,
                    name=row.name,
                    domain=row.domain,
                    system_id=row.system_id,
                    data_uses=[],
                    asset_type="Cookie",
                )
            continue

        if row.pud_system_id:
            system_id = row.pud_system_id
            data_uses = [row.data_use]
        else:
            system_id = row.system_id
            data_uses = []

        # Asset unique identifier is a composite of name, asset_type, domain, base_url, system_id. All assets here are cookies.
        identifier = f"{row.name}_{'Cookie'}_{row.domain}_{row.path}_{system_id}"
        if identifier in assets_to_create.keys():
            # If the asset already exists, append the data_use to the existing asset
            assets_to_create[identifier].data_uses.extend(data_uses)
        else:
            # If the asset does not exist, create a new asset with the data_use
            assets_to_create[identifier] = Asset(
                id=str(uuid.uuid4()),
                created_at=row.created_at,
                updated_at=row.updated_at,
                name=row.name,
                domain=row.domain,
                system_id=system_id,
                data_uses=data_uses,
                asset_type="Cookie",
            )

    # Insert the assets into the asset table
    assets_list = [
        {
            "id": asset.id if asset.id else str(uuid.uuid4()),
            "created_at": asset.created_at,
            "updated_at": asset.updated_at,
            "name": asset.name,
            "domain": asset.domain,
            "system_id": asset.system_id,
            "data_uses": asset.data_uses,
            "asset_type": "Cookie",
        }
        for asset in assets_to_create.values()
    ]

    if assets_list:
        logger.debug("Inserting assets into asset table ")
        connection.execute(
            sa.text(
                "INSERT INTO asset (id, created_at, updated_at, name, domain, system_id, data_uses, asset_type, with_consent) "
                "VALUES (:id, :created_at, :updated_at, :name, :domain, :system_id, :data_uses, :asset_type, false) "
                "ON CONFLICT DO NOTHING"
            ),
            assets_list,
        )
    else:
        logger.debug("No assets to insert into asset table. Skipping.")

    # Delete the cookies table
    logger.debug("Deleting cookies table")
    connection.execute(sa.text("DROP TABLE cookies"))


def downgrade():
    # Recreate the cookies table
    logger.debug("Recreating cookies table")
    op.create_table(
        "cookies",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("domain", sa.String(), nullable=True),
        sa.Column("path", sa.String(), nullable=True),
        sa.Column("system_id", sa.String(), nullable=True),
        sa.Column("privacy_declaration_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["privacy_declaration_id"], ["privacydeclaration.id"], ondelete="SET NULL"
        ),
        sa.ForeignKeyConstraint(["system_id"], ["ctl_systems.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "privacy_declaration_id", name="_cookie_name_privacy_declaration_uc"
        ),
    )
    op.create_index(op.f("ix_cookies_id"), "cookies", ["id"], unique=False)
    op.create_index(op.f("ix_cookies_name"), "cookies", ["name"], unique=False)
    op.create_index(
        op.f("ix_cookies_privacy_declaration_id"),
        "cookies",
        ["privacy_declaration_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_cookies_system_id"), "cookies", ["system_id"], unique=False
    )
    # ### end Alembic commands ###

    connection = op.get_bind()
    result = connection.execute(
        sa.text(
            "SELECT id, name, domain, system_id, data_uses FROM asset WHERE asset_type = 'Cookie'"
        )
    )

    # fetch all privacy declarations
    privacy_declarations = connection.execute(
        sa.text("SELECT id, system_id, data_use FROM privacydeclaration")
    ).fetchall()
    # create a mapping of system_id and data_use to privacy_declaration_id
    privacy_declaration_mapping: Dict[Tuple[str, str], str] = {}
    for pud in privacy_declarations:
        privacy_declaration_mapping[(pud.system_id, pud.data_use)] = pud.id

    logger.debug("Migrating existing assets to cookies")
    for row in result:
        # Find the privacy declaration ID matching the system ID and data use
        privacy_declaration_ids = []
        for data_use in row.data_uses:
            pud_id = privacy_declaration_mapping.get((row.system_id, data_use), None)
            if pud_id:
                privacy_declaration_ids.append(pud_id)

        cookies_to_insert = []

        if len(privacy_declaration_ids) == 0:
            # If no privacy declaration match we attach to system_id
            cookies_to_insert.append(
                {
                    "id": row.id,
                    "name": row.name,
                    "domain": row.domain,
                    "system_id": row.system_id,
                    "privacy_declaration_id": None,
                }
            )
        else:
            for privacy_declaration_id in privacy_declaration_ids:
                # Create a cookie for each privacy declaration ID
                # generate a new ID
                cookie_id = str(uuid.uuid4())
                cookies_to_insert.append(
                    {
                        "id": cookie_id,
                        "name": row.name,
                        "domain": row.domain,
                        "system_id": None,
                        "privacy_declaration_id": privacy_declaration_id,
                    }
                )

        if cookies_to_insert:
            connection.execute(
                sa.text(
                    "INSERT INTO cookies (id, created_at, updated_at, name, domain, system_id, privacy_declaration_id) "
                    "VALUES (:id, NOW(), NOW(), :name, :domain, :system_id, :privacy_declaration_id) "
                    "ON CONFLICT DO NOTHING"
                ),
                cookies_to_insert,
            )



"""requesttask

Revision ID: 55bedede956d
Revises: 6cfd59e7920a
Create Date: 2024-04-04 04:12:25.332952

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "55bedede956d"
down_revision = "6cfd59e7920a"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "requesttask",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("privacy_request_id", sa.String(), nullable=True),
        sa.Column("collection_address", sa.String(), nullable=False),
        sa.Column("dataset_name", sa.String(), nullable=False),
        sa.Column("collection_name", sa.String(), nullable=False),
        sa.Column("action_type", sa.String(), nullable=False),
        sa.Column("status", sa.String(), nullable=False),
        sa.Column(
            "upstream_tasks", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column(
            "downstream_tasks", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column(
            "all_descendant_tasks",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
        sa.Column(
            "access_data",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "data_for_erasures",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("rows_masked", sa.Integer(), nullable=True),
        sa.Column("consent_sent", sa.Boolean(), nullable=True),
        sa.Column("collection", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column(
            "traversal_details", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["privacy_request_id"], ["privacyrequest.id"], ondelete="SET NULL"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_requesttask_action_type"), "requesttask", ["action_type"], unique=False
    )
    op.create_index(
        op.f("ix_requesttask_collection_address"),
        "requesttask",
        ["collection_address"],
        unique=False,
    )
    op.create_index(
        op.f("ix_requesttask_collection_name"),
        "requesttask",
        ["collection_name"],
        unique=False,
    )
    op.create_index(
        op.f("ix_requesttask_dataset_name"),
        "requesttask",
        ["dataset_name"],
        unique=False,
    )
    op.create_index(op.f("ix_requesttask_id"), "requesttask", ["id"], unique=False)
    op.create_index(
        op.f("ix_requesttask_privacy_request_id"),
        "requesttask",
        ["privacy_request_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_requesttask_status"), "requesttask", ["status"], unique=False
    )
    op.add_column(
        "privacyrequest",
        sa.Column(
            "filtered_final_upload",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyrequest",
        sa.Column(
            "access_result_urls",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column("privacyrequest", "access_result_urls")
    op.drop_column("privacyrequest", "filtered_final_upload")
    op.drop_index(op.f("ix_requesttask_status"), table_name="requesttask")
    op.drop_index(op.f("ix_requesttask_privacy_request_id"), table_name="requesttask")
    op.drop_index(op.f("ix_requesttask_id"), table_name="requesttask")
    op.drop_index(op.f("ix_requesttask_dataset_name"), table_name="requesttask")
    op.drop_index(op.f("ix_requesttask_collection_name"), table_name="requesttask")
    op.drop_index(op.f("ix_requesttask_collection_address"), table_name="requesttask")
    op.drop_index(op.f("ix_requesttask_action_type"), table_name="requesttask")
    op.drop_table("requesttask")



"""
Helpers for estimating the size of large collections of access data.
"""

from __future__ import annotations

import json
import sys
from typing import List, Optional

from loguru import logger

from fides.api.util.collection_util import Row
from fides.api.util.custom_json_encoder import CustomJSONEncoder

# 640MB threshold for external storage
# We only generate an estimated size for large datasets so we want to be conservative
# and fallback to external storage even if we haven't hit the 1GB max limit.
# We also want to pad for encryption and base64 encoding.
LARGE_DATA_THRESHOLD_BYTES = 640 * 1024 * 1024  # 640MB


def calculate_data_size(data: List[Row]) -> int:  # noqa: D401 – utility function
    """Return an approximate JSON-serialized size (in bytes) for a list of *Row*.

    The implementation purposefully avoids serializing the entire payload when
    *data* is large.  For collections >1000 rows we sample a subset, measure the
    encoded size, then extrapolate.  This keeps memory usage bounded while still
    giving us an order-of-magnitude estimate suitable for "should I stream this
    out to S3?" decisions.
    """

    if not data:
        return 0

    try:
        data_count = len(data)

        # For very large datasets, estimate size from a sample to avoid memory issues
        if data_count > 1000:
            logger.debug(
                f"Calculating size for large dataset ({data_count} rows) using sampling"
            )

            sample_size = min(500, max(100, data_count // 20))  # 5 % capped at 500

            # stratified sampling – take items spaced across the set when possible
            if data_count > sample_size * 3:
                step = data_count // sample_size
                sample_indices = list(range(0, data_count, step))[:sample_size]
                sample = [data[i] for i in sample_indices]
            else:
                sample = data[:sample_size]

            sample_json = json.dumps(
                sample, cls=CustomJSONEncoder, separators=(",", ":")
            )
            sample_bytes = len(sample_json.encode("utf-8"))

            avg_record_size = sample_bytes / sample_size
            content_size = int(avg_record_size * data_count)

            # overhead: 2 bytes for [] plus a comma between every record plus 1 % slack
            structure_overhead = 2 + (data_count - 1) + int(content_size * 0.01)
            return content_size + structure_overhead

        # small datasets – just measure
        json_str = json.dumps(data, cls=CustomJSONEncoder, separators=(",", ":"))
        return len(json_str.encode("utf-8"))

    except (TypeError, ValueError) as exc:
        logger.warning(
            f"Failed to calculate JSON size, falling back to sys.getsizeof: {exc}"
        )
        return sys.getsizeof(data)


def is_large_data(
    data: List[Row], threshold_bytes: Optional[int] = None
) -> bool:  # noqa: D401
    """Return *True* if *data* is likely to exceed *threshold_bytes* when serialized."""

    if not data:
        return False

    threshold = (
        threshold_bytes if threshold_bytes is not None else LARGE_DATA_THRESHOLD_BYTES
    )
    size = calculate_data_size(data)
    if size > threshold:
        logger.info(
            f"Data size ({size:,} bytes) exceeds threshold ({threshold:,} bytes) – using external storage"
        )
        return True
    return False


__all__ = [
    "calculate_data_size",
    "is_large_data",
    "LARGE_DATA_THRESHOLD_BYTES",
]



from typing import TYPE_CHECKING, Any, Optional, cast

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    UniqueConstraint,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql import func

from fides.api.db.base_class import Base, FidesBase
from fides.api.db.util import EnumColumn
from fides.api.models.manual_tasks.manual_task_log import ManualTaskLog
from fides.api.schemas.manual_tasks.manual_task_config import (
    ManualTaskConfigurationType,
    ManualTaskFieldMetadata,
    ManualTaskFieldType,
)
from fides.api.schemas.manual_tasks.manual_task_schemas import (
    ManualTaskExecutionTiming,
    ManualTaskLogStatus,
)

if TYPE_CHECKING:  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task import ManualTask  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task_instance import (
        ManualTaskInstance,  # pragma: no cover
    )
    from fides.api.models.manual_tasks.manual_task_instance import (
        ManualTaskSubmission,  # pragma: no cover
    )


class ManualTaskConfig(Base):
    """Model for storing manual task configurations.
    A single configuration may have many fields of different types.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        return "manual_task_config"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_config_task_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    config_type = Column(EnumColumn(ManualTaskConfigurationType), nullable=False)
    version = Column(Integer, nullable=False, default=1)
    is_current = Column(Boolean, nullable=False, default=True)
    execution_timing = Column(
        EnumColumn(ManualTaskExecutionTiming),
        nullable=False,
        default=ManualTaskExecutionTiming.pre_execution,
    )

    __table_args__ = (
        Index("ix_manual_task_config_config_type", "config_type"),
        Index("ix_manual_task_config_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="configs", viewonly=True)
    instances = relationship(
        "ManualTaskInstance", back_populates="config", uselist=True, viewonly=True
    )
    submissions = relationship(
        "ManualTaskSubmission", back_populates="config", uselist=True, viewonly=True
    )
    field_definitions = relationship(
        "ManualTaskConfigField",
        back_populates="config",
        cascade="all, delete-orphan",
        uselist=True,
    )
    logs = relationship(
        "ManualTaskLog",
        back_populates="config",
        primaryjoin="ManualTaskConfig.id == ManualTaskLog.config_id",
        cascade="all, delete-orphan",
    )

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = True
    ) -> "ManualTaskConfig":
        """Create a new manual task configuration."""
        # Validate config_type
        try:
            ManualTaskConfigurationType(data["config_type"])
        except ValueError:
            raise ValueError(f"Invalid config type: {data['config_type']}")

        config = super().create(db=db, data=data, check_name=check_name)

        # Log the config creation as a task-level log
        ManualTaskLog.create_log(
            db=db,
            task_id=data["task_id"],
            config_id=config.id,
            status=ManualTaskLogStatus.created,
            message=f"Created manual task configuration for {data['config_type']}",
            details={
                "config_type": data["config_type"],
            },
        )
        return config

    def get_field(self, field_key: str) -> Optional["ManualTaskConfigField"]:
        """Get a field by its key."""
        for field in self.field_definitions:
            if field.field_key == field_key:
                return field
        return None


class ManualTaskConfigField(Base):
    """Model for storing fields associated with each config."""

    @declared_attr
    def __tablename__(cls) -> str:
        return "manual_task_config_field"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_config_field_task_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    config_id = Column(
        String,
        ForeignKey(
            "manual_task_config.id",
            name="manual_task_config_field_config_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    field_key = Column(String, nullable=False)
    field_type = Column(
        EnumColumn(ManualTaskFieldType), nullable=False
    )  # Using ManualTaskFieldType
    field_metadata: dict[str, Any] = cast(
        dict[str, Any], Column(JSONB, nullable=False, default={})
    )

    __table_args__ = (
        UniqueConstraint("config_id", "field_key", name="unique_field_key_per_config"),
        Index("ix_manual_task_config_field_config_id", "config_id"),
        Index("ix_manual_task_config_field_field_key", "field_key"),
        Index("ix_manual_task_config_field_task_id", "task_id"),
    )

    # Relationships
    config = relationship(
        "ManualTaskConfig", back_populates="field_definitions", viewonly=True
    )
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="field",
        uselist=True,
        cascade="all, delete-orphan",
    )

    @property
    def field_metadata_model(self) -> ManualTaskFieldMetadata:
        """Get the field metadata as a Pydantic model."""
        assert isinstance(
            self.field_metadata, dict
        ), "field_metadata must be a dictionary"
        return ManualTaskFieldMetadata.model_validate(self.field_metadata)

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = True
    ) -> "ManualTaskConfigField":
        """Create a new manual task config field."""
        # Get the config to access its task_id and check if it exists
        try:
            config = (
                db.query(ManualTaskConfig)
                .filter(ManualTaskConfig.id == data["config_id"])
                .first()
            )
            if not config:
                raise ValueError(f"Config with id {data['config_id']} not found")
        except Exception as e:
            raise ValueError(f"Config with id {data['config_id']} not found: {e}")

        # Create the field and let SQLAlchemy complex type validation handled in service.
        field = super().create(db=db, data=data, check_name=check_name)

        # Create a log entry
        ManualTaskLog.create_log(
            db=db,
            task_id=config.task_id,
            config_id=data["config_id"],
            status=ManualTaskLogStatus.created,
            message=f"Created manual task config field for {data['field_key']}",
        )
        return field



"""
Entrypoint for the Fides command-line.
"""

# pylint: disable=wrong-import-position
import warnings

# Ignore the UserWarning from the Snowflake module to keep CLI output clean
warnings.filterwarnings("ignore", category=UserWarning, module="snowflake")

from importlib.metadata import version
from platform import system

from fideslog.sdk.python.client import AnalyticsClient
from rich_click import Context, echo, group, option, pass_context, secho, version_option

import fides
from fides.cli.utils import check_server
from fides.config import get_config

from . import cli_formatting
from .commands.annotate import annotate
from .commands.db import database
from .commands.deploy import deploy
from .commands.generate import generate
from .commands.pull import pull
from .commands.scan import scan
from .commands.ungrouped import (
    delete,
    evaluate,
    get_resource,
    init,
    list_resources,
    parse,
    push,
    status,
    webserver,
    worker,
)
from .commands.user import user
from .commands.view import view
from .exceptions import LocalModeException

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
LOCAL_COMMANDS = [deploy, evaluate, generate, init, scan, parse, view, webserver]
LOCAL_COMMAND_NAMES = {command.name for command in LOCAL_COMMANDS}
API_COMMANDS = [
    annotate,
    database,
    delete,
    get_resource,
    list_resources,
    status,
    pull,
    push,
    worker,
    user,
]
ALL_COMMANDS = API_COMMANDS + LOCAL_COMMANDS
VERSION = fides.__version__
APP = fides.__name__
PACKAGE = "ethyca-fides"


@group(  # type: ignore
    context_settings=CONTEXT_SETTINGS,
    invoke_without_command=True,
    name="fides",
)
@version_option(version=VERSION)
@option(
    "--config-path",
    "-f",
    "config_path",
    show_default=True,
    help="Path to a Fides config file. _Defaults to `.fides/fides.toml`._",
)
@option(
    "--local",
    is_flag=True,
    help="Run in `local_mode`. Where possible, this will force commands to run without the need for a server.",
)
@pass_context
def cli(ctx: Context, config_path: str, local: bool) -> None:
    """
    __Command-line tool for the Fides privacy engineering platform.__

    ---

    _Note: The common MANIFESTS_DIR argument _always_ defaults to ".fides/" if not specified._
    """

    ctx.ensure_object(dict)
    config = get_config(config_path, verbose=True)
    command = ctx.invoked_subcommand or ""

    if not (local or config.cli.local_mode):
        config.cli.local_mode = False
    else:
        config.cli.local_mode = True

    if config.cli.local_mode and command not in LOCAL_COMMAND_NAMES:
        raise LocalModeException(command)

    # Run the help command if no subcommand is passed
    if not command:
        echo(cli.get_help(ctx))

    # Analytics requires explicit opt-in
    no_analytics = config.user.analytics_opt_out
    if not no_analytics:
        ctx.meta["ANALYTICS_CLIENT"] = AnalyticsClient(
            client_id=config.cli.analytics_id,
            developer_mode=config.test_mode,
            os=system(),
            product_name=APP + "-cli",
            production_version=version(PACKAGE),
        )

    # Setting the config context after all mutations
    ctx.obj["CONFIG"] = config


# Add all commands here before dynamically checking them in the CLI
for cli_command in ALL_COMMANDS:
    cli.add_command(cli_command)



"""remove deprecated data uses for fideslang 1.4

Revision ID: 5307999c0dac
Revises: 76c02f99eec1
Create Date: 2023-06-11 11:15:53.386526

"""

from alembic import op
from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "5307999c0dac"
down_revision = "76c02f99eec1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    logger.info("Removing default data uses.")
    bind: Connection = op.get_bind()
    bind.execute(text("DELETE FROM ctl_data_uses WHERE is_default = TRUE;"))

    # This leaves the data uses empty, but the step to seed them back comes _after_ these migrations


def downgrade() -> None:
    pass
    # This is not reversible, because it relies on an older version of Fideslang



from typing import List, Optional

from pydantic import ConfigDict, field_validator

from fides.api.oauth.roles import RoleRegistryEnum
from fides.api.schemas.base_class import FidesSchema
from fides.common.api.scope_registry import SCOPE_REGISTRY, ScopeRegistryEnum


class UserPermissionsCreate(FidesSchema):
    """Data required to create a FidesUserPermissions record

    Users will be assigned role(s) directly which are associated with a list of scopes. Scopes
    cannot be assigned directly to users.
    """

    roles: List[RoleRegistryEnum]
    model_config = ConfigDict(use_enum_values=True)


class UserPermissionsEdit(UserPermissionsCreate):
    """Data required to edit a FidesUserPermissions record."""

    id: Optional[str] = (
        None  # I don't think this should be in the request body, so making it optional.
    )


class UserPermissionsResponse(UserPermissionsCreate):
    """Response after creating, editing, or retrieving a FidesUserPermissions record."""

    id: str
    user_id: str
    total_scopes: List[
        ScopeRegistryEnum
    ]  # Returns a list of scopes inherited via roles
    model_config = ConfigDict(use_enum_values=True)

    @field_validator("total_scopes", mode="before")
    @classmethod
    def validate_obsolete_total_scopes(
        cls, total_scopes: List[ScopeRegistryEnum]
    ) -> List[ScopeRegistryEnum]:
        """Filter out obsolete total scopes if the scope registry has changed"""
        return [scope for scope in total_scopes or [] if scope in SCOPE_REGISTRY]



from enum import Enum as EnumType
from typing import Dict, List, Set, Type

from fideslang.default_taxonomy import DEFAULT_TAXONOMY
from fideslang.validation import FidesKey
from sqlalchemy import func, select, text
from sqlalchemy.orm import Session

from fides.api import common_exceptions

from fides.api.models.sql_models import (  # type: ignore[attr-defined] # isort: skip
    DataCategory as DataCategoryDbModel,
    Dataset,
)


def generate_fides_data_categories() -> Type[EnumType]:
    """Programmatically generated the DataCategory enum based on the imported Fides data."""
    FidesDataCategory = EnumType(  # type: ignore
        "FidesDataCategory",
        {
            cat.fides_key: cat.fides_key
            for cat in DEFAULT_TAXONOMY.data_category  # pylint:disable=not-an-iterable
        },  # pylint:disable=not-an-iterable
    )
    return FidesDataCategory


# TODO: Move away from using this, it conflicts with the DataCategory model
DataCategory = generate_fides_data_categories()


def get_data_categories_from_db(db: Session) -> List[FidesKey]:
    """Query for existing data categories in the db using a synchronous session"""
    return [cat[0] for cat in db.query(DataCategoryDbModel.fides_key).all()]


def _validate_data_category(
    db: Session,
    data_category: str,
) -> str:
    """Checks that the data category passed in is currently supported."""
    valid_categories = get_data_categories_from_db(db=db)
    if data_category not in valid_categories:
        raise common_exceptions.DataCategoryNotSupported(
            f"The data category '{data_category}' was not found in the database, and is therefore not valid for use here."
        )
    return data_category


def get_user_data_categories() -> List[str]:
    # organizations need to be extra careful about how these are used -
    # especially for erasure! Therefore, a safe default for "out of the
    # box" behaviour is to exclude these
    excluded_data_categories = [
        "user.financial",
        "user.payment",
        "user.authorization",
    ]
    all_data_categories = [
        str(category.fides_key)
        for category in DEFAULT_TAXONOMY.data_category  # pylint:disable=not-an-iterable
    ]
    return filter_data_categories(all_data_categories, excluded_data_categories)


def filter_data_categories(
    categories: List[str], excluded_categories: List[str]
) -> List[str]:
    """
    Filter data categories and their children out of a list of categories.

    We only want user-related data categories, but not the parent category
    We also only want 2nd level categories, otherwise there are policy conflicts
    """
    user_categories = [
        category
        for category in categories
        if category.startswith("user.") and len(category.split(".")) < 3
    ]
    if excluded_categories:
        duplicated_categories = [
            category
            for excluded_category in excluded_categories
            for category in user_categories
            if not category.startswith(excluded_category)
        ]
        default_categories = {
            category
            for category in duplicated_categories
            if duplicated_categories.count(category) == len(excluded_categories)
        }
        return sorted(list(default_categories))
    return sorted(user_categories)


def get_data_categories_map(db: Session) -> Dict[str, Set[str]]:
    """
    Returns a map of all datasets, where the keys are the fides keys
    of each dataset and the value is a set of data categories associated with each dataset
    """

    subquery = (
        select(
            Dataset.fides_key,
            func.jsonb_array_elements_text(
                text(
                    "jsonb_path_query(collections::jsonb, '$.** ? (@.data_categories != null).data_categories')"
                )
            ).label("category"),
        ).select_from(Dataset)
    ).cte()

    query = (
        select(
            [subquery.c.fides_key, func.array_agg(func.distinct(subquery.c.category))]
        )
        .select_from(subquery)
        .group_by(subquery.c.fides_key)
    )
    result = db.execute(query)
    return {key: set(value) if value else set() for key, value in result.all()}



"""add tags as part of fideslang 1.1.0

Revision ID: d8f8bd52754f
Revises: be432bd23596
Create Date: 2022-07-07 02:54:13.724757

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d8f8bd52754f"
down_revision = "be432bd23596"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "data_categories",
        sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True),
    )
    op.add_column(
        "data_qualifiers",
        sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True),
    )
    op.add_column(
        "data_subjects",
        sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True),
    )
    op.add_column(
        "data_uses", sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True)
    )
    op.add_column(
        "datasets", sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True)
    )
    op.add_column(
        "organizations",
        sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True),
    )
    op.add_column(
        "policies", sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True)
    )
    op.add_column(
        "registries", sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True)
    )
    op.add_column(
        "systems", sa.Column("tags", postgresql.ARRAY(sa.VARCHAR()), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("systems", "tags")
    op.drop_column("registries", "tags")
    op.drop_column("policies", "tags")
    op.drop_column("organizations", "tags")
    op.drop_column("datasets", "tags")
    op.drop_column("data_uses", "tags")
    op.drop_column("data_subjects", "tags")
    op.drop_column("data_qualifiers", "tags")
    op.drop_column("data_categories", "tags")
    # ### end Alembic commands ###



import time as time_module
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Tuple

from loguru import logger

from fides.api.models.attachment import Attachment, AttachmentType
from fides.api.schemas.storage.storage import StorageDetails


@dataclass
class AttachmentData:
    """Data structure for attachment metadata and content.
    Using a dataclass rather than a Pydantic model here for the following reasons:
    - The data structure is simple and doesn't need complex validation.
    - The fields being used have already been validated and are properly typed.
    - The class is used internally for data transfer, not for API serialization.
    - Performance is important since this is used in a data processing pipeline.
    """

    file_name: str
    file_size: Optional[int]
    download_url: Optional[str]
    content_type: str
    bucket_name: str
    file_key: str
    storage_key: str

    def to_upload_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for upload, including presigned URL."""
        return {
            "file_name": self.file_name,
            "file_size": self.file_size,
            "download_url": self.download_url,
            "content_type": self.content_type,
        }

    def to_storage_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for storage, including the elements needed to recreated the presigned URL."""
        return {
            "file_name": self.file_name,
            "file_size": self.file_size,
            "content_type": self.content_type,
            "bucket_name": self.bucket_name,
            "file_key": self.file_key,
            "storage_key": self.storage_key,
        }


def get_attachments_content(
    loaded_attachments: List[Attachment],
) -> Iterator[AttachmentData]:
    """
    Retrieves all attachments associated with a privacy request that are marked to be included with the access package.
    Yields AttachmentData objects containing attachment metadata and download urls.
    Uses generators to minimize memory usage.

    Args:
        loaded_attachments: List of Attachment objects to process

    Yields:
        AttachmentData object containing attachment metadata and url
    """
    start_time = time_module.time()
    processed_count = 0
    skipped_count = 0
    error_count = 0
    total_size = 0

    for attachment in loaded_attachments:
        if attachment.attachment_type != AttachmentType.include_with_access_package:
            skipped_count += 1
            continue

        try:
            # Get size and download URL using retrieve_attachment
            size, url = attachment.retrieve_attachment()
            total_size += size if size else 0
            if url is None:
                logger.warning(
                    "No download URL retrieved for attachment {}", attachment.file_name
                )
                skipped_count += 1
                continue

            processed_count += 1
            yield AttachmentData(
                file_name=attachment.file_name,
                file_size=size,
                download_url=str(url) if url else None,
                content_type=attachment.content_type,
                bucket_name=attachment.config.details[StorageDetails.BUCKET.value],
                file_key=attachment.file_key,
                storage_key=attachment.storage_key,
            )

        except Exception as e:
            error_count += 1
            logger.error(
                "Error processing attachment {}: {}", attachment.file_name, str(e)
            )
            continue

    # Log final metrics
    time_taken = time_module.time() - start_time
    logger.bind(
        time_to_process=time_taken,
        total_attachments=len(loaded_attachments),
        processed_attachments=processed_count,
        skipped_attachments=skipped_count,
        error_attachments=error_count,
        total_size_bytes=total_size,
    ).info("Attachment processing complete")


def process_attachments_for_upload(
    attachments: Iterator[AttachmentData],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Process attachments into separate upload and storage formats.
    Returns both formats:
    - upload_attachments: Used for uploading to access packages
    - storage_attachments: Used for saving filtered access results
    """
    upload_attachments = []
    storage_attachments = []

    for attachment in attachments:
        storage_attachments.append(attachment.to_storage_dict())
        upload_attachments.append(attachment.to_upload_dict())

    return upload_attachments, storage_attachments



from enum import Enum
from typing import ByteString, Optional

from sqlalchemy import Column, Index, String
from sqlalchemy.dialects.postgresql import BYTEA
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import flag_modified

from fides.api.db.base_class import Base


class DBCacheNamespace(Enum):
    """Namespaces for the DBCache"""

    LIST_PRIVACY_EXPERIENCE = "list-privacy-experience"


class DBCache(Base):
    """
    Cache table for storing arbitrary data, useful when in-memory caches aren't enough.
    For example if cache contents need to be persisted across server restarts,
    or if the cache needs to be shared across different Fides instances.

    Warning: Cache contents are NOT encrypted, this shouldn't be used for storing
    any sort of personal data or sensitive information.
    """

    namespace = Column(
        String, nullable=False, index=True
    )  # Add a namespace since the same cache key could technically be used for different contexts
    cache_key = Column(String, nullable=False)
    cache_value = Column(BYTEA, nullable=False)

    __table_args__ = (
        Index("ix_dbcache_namespace_cache_key", "namespace", "cache_key", unique=True),
    )

    @classmethod
    def get_cache_entry(
        cls,
        db: Session,
        namespace: DBCacheNamespace,
        cache_key: str,
    ) -> Optional["DBCache"]:
        """
        Retrieves the cache entry for the given cache_key
        """
        return (
            db.query(cls)
            .filter(cls.namespace == namespace.value, cls.cache_key == cache_key)
            .first()
        )

    @classmethod
    def get_cache_value(
        cls,
        db: Session,
        namespace: DBCacheNamespace,
        cache_key: str,
    ) -> Optional[ByteString]:
        """
        Retrieves the cache value for the given cache_key
        """
        cache_entry = cls.get_cache_entry(db, namespace, cache_key)

        return cache_entry.cache_value if cache_entry else None

    @classmethod
    def set_cache_value(
        cls,
        db: Session,
        namespace: DBCacheNamespace,
        cache_key: str,
        cache_value: ByteString,
    ) -> "DBCache":
        """
        Upserts the cache value for the given cache_key
        """
        db_cache_entry = cls.get_cache_entry(db, namespace, cache_key)
        if db_cache_entry:
            db_cache_entry.cache_value = cache_value
            # We manually flag it as modified so that the update runs even if the cache_value hasn't changed
            # so the updated_at field of the cache entry gets updated.
            flag_modified(db_cache_entry, "cache_value")
        else:
            db_cache_entry = cls(
                namespace=namespace.value, cache_key=cache_key, cache_value=cache_value
            )

        db.add(db_cache_entry)
        db.commit()
        db.refresh(db_cache_entry)
        return db_cache_entry

    @classmethod
    def delete_cache_entry(
        cls,
        db: Session,
        namespace: DBCacheNamespace,
        cache_key: str,
    ) -> None:
        """
        Deletes the cache entry for the given cache_key
        """
        db.query(cls).filter(
            cls.namespace == namespace.value, cls.cache_key == cache_key
        ).delete()
        db.commit()

    @classmethod
    def clear_cache_for_namespace(
        cls,
        db: Session,
        namespace: DBCacheNamespace,
    ) -> None:
        """
        Deletes all cache entries for the given namespace
        """
        db.query(cls).filter(cls.namespace == namespace.value).delete()
        db.commit()

    @classmethod
    def clear_cache(
        cls,
        db: Session,
    ) -> None:
        """
        Deletes all cache entries
        """
        db.query(cls).delete()
        db.commit()
